TEST_VC_DOCUMENT_NESTED_PARTIAL_PROOF_BBS = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "id": "https://example.gov/credentials/3732",
    "type": ["UniversityDegreeCredential", "VerifiableCredential"],
    "credentialSubject": {
        "id": "did:example:489398593",
        "degree": {
            "id": "urn:bnid:_:c14n0",
            "type": "BachelorDegree",
            "name": "Bachelor of Science and Arts",
        },
    },
    "issuanceDate": "2020-03-10T04:24:12.164Z",
    "issuer": "did:example:489398593",
    "proof": {
        "type": "BbsBlsSignatureProof2020",
        "created": "2021-04-26T17:55:05Z",
        "nonce": "58pQbOsBR7l+FwAGs1bDWiCyQXJFkmSNBHC0BDzX4BU6khM6VbJZBqb5RGONkGFJBNQ=",
        "proofPurpose": "assertionMethod",
        "proofValue": "AA4f74GNgyGujm0aHzKd2xuUpetxhpFaXhW6twInMfmFNZYW2/RlShvRGE6Ios/dgr5e+qyGGOGkBWSdSOaFGsseo6Ne/0429iKAZpWo/abD6r8kIszn8L5UL43NxCia/UtQB5HjqBPYelAsepN3aNryZNUc0h7UcyjTqgp9wGfPADW4KJIBXv4SvwqwRZgcDzdKxwAAAHSStf9LAUTsKXnlWL7hE0x86/4wzpDFRWCE/WgkCz/mqQlt+vlWDwm9k7PlwiPoLhYAAAACBLS2axWssG0oNnqk02If5qm6ToalAP/TGB/cORH00KNrKMvhMMwI54odGP5qy4qJmQp+cYcj4bffIfaXbRGNqJdHYKbeHk5ZV/A91vnI3mxQfhtKp1YmNC2DpUfw/hSTAubh5Iv2UMFWc1Gl+hYuDAAAAARiA7M07jBRe+QGbPgLg1nFgiD26GDBD+wYm93qM+HfwBKYqEbeXVveTsx589OTilT5HP2b6cOfSwCJyU8xU/uKPNYhNGqkGZgPXjIVfvRWFblSX822hnMBOYzdCroarsdUCom4NMlFvAQeZAQ6CtZUkXSKAizYCdx5yeHjjkveAg==",
        "verificationMethod": "did:example:489398593#test",
    },
}



from unittest import IsolatedAsyncioTestCase

from aries_askar import AskarError

from acapy_agent.askar.profile import AskarProfileSession
from acapy_agent.did.indy.indy_manager import DidIndyManager
from acapy_agent.tests import mock
from acapy_agent.utils.testing import create_test_profile
from acapy_agent.wallet.did_method import DIDMethods
from acapy_agent.wallet.error import WalletError
from acapy_agent.wallet.key_type import KeyTypes


class TestIndyManager(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(
            DIDMethods, mock.MagicMock(DIDMethods, auto_spec=True)
        )
        self.profile.context.injector.bind_instance(KeyTypes, KeyTypes())

    def test_init(self):
        assert DidIndyManager(self.profile)

    @mock.patch.object(AskarProfileSession, "handle")
    async def test_register(self, mock_handle):
        mock_handle.insert_key = mock.CoroutineMock()
        mock_handle.insert = mock.CoroutineMock()
        manager = DidIndyManager(self.profile)
        result = await manager.register({})
        assert result.get("did")
        assert result.get("verkey")
        mock_handle.insert_key.assert_called_once()
        mock_handle.insert.assert_called_once()

        # error saving key
        mock_handle.insert_key.side_effect = AskarError(
            code=1, message="Error saving key"
        )
        with self.assertRaises(WalletError):
            await manager.register({})

    @mock.patch.object(AskarProfileSession, "handle")
    async def test_register_with_seed_with_key_type(self, mock_handle):
        mock_handle.insert_key = mock.CoroutineMock()
        mock_handle.insert = mock.CoroutineMock()
        manager = DidIndyManager(self.profile)

        result = await manager.register({"key_type": "ed25519"})
        assert result.get("did")
        assert result.get("verkey")

    @mock.patch.object(AskarProfileSession, "handle")
    async def test_register_with_seed_with_defined_did(self, mock_handle):
        mock_handle.insert_key = mock.CoroutineMock()
        mock_handle.insert = mock.CoroutineMock()
        manager = DidIndyManager(self.profile)

        result = await manager.register({"did": "did:indy:WRfXPg8dantKVubE3HX8pw"})
        assert result.get("did") == "did:indy:WRfXPg8dantKVubE3HX8pw"
        assert result.get("verkey")

    @mock.patch.object(AskarProfileSession, "handle")
    async def test_register_with_seed_with_all_options(self, mock_handle):
        self.profile.settings.set_value("wallet.allow_insecure_seed", True)
        mock_handle.insert_key = mock.CoroutineMock()
        mock_handle.insert = mock.CoroutineMock()
        manager = DidIndyManager(self.profile)

        result = await manager.register(
            {
                "did": "did:indy:WRfXPg8dantKVubE3HX8pw",
                "key_type": "ed25519",
                "seed": "000000000000000000000000Trustee1",
            }
        )
        assert result.get("did") == "did:indy:WRfXPg8dantKVubE3HX8pw"
        assert result.get("verkey")



"""Admin routes for presentations."""

from typing import Mapping, Optional

from marshmallow import EXCLUDE, fields

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.models.openapi import OpenAPISchema
from ...messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_CRED_REV_ID_EXAMPLE,
    INDY_CRED_REV_ID_VALIDATE,
    INDY_REV_REG_ID_EXAMPLE,
    INDY_REV_REG_ID_VALIDATE,
    INDY_SCHEMA_ID_EXAMPLE,
    INDY_SCHEMA_ID_VALIDATE,
    UUID4_EXAMPLE,
)
from .non_rev_interval import IndyNonRevocationIntervalSchema


class IndyCredInfo(BaseModel):
    """Indy cred info, as holder gets via indy-sdk."""

    class Meta:
        """IndyCredInfo metadata."""

        schema_class = "IndyCredInfoSchema"

    def __init__(
        self,
        referent: Optional[str] = None,
        attrs: Optional[Mapping] = None,
        schema_id: Optional[str] = None,
        cred_def_id: Optional[str] = None,
        rev_reg_id: Optional[str] = None,
        cred_rev_id: Optional[str] = None,
    ):
        """Initialize indy cred info."""
        self.referent = referent
        self.attrs = attrs
        self.schema_id = schema_id
        self.cred_def_id = cred_def_id
        self.rev_reg_id = rev_reg_id
        self.cred_rev_id = cred_rev_id


class IndyCredInfoSchema(BaseModelSchema):
    """Schema for indy cred-info."""

    class Meta:
        """Schema metadata."""

        model_class = IndyCredInfo
        unknown = EXCLUDE

    referent = fields.Str(
        metadata={"description": "Wallet referent", "example": UUID4_EXAMPLE}
    )
    attrs = fields.Dict(
        keys=fields.Str(metadata={"example": "userid"}),
        values=fields.Str(metadata={"example": "alice"}),
        metadata={"description": "Attribute names and value"},
    )
    schema_id = fields.Str(
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    cred_def_id = fields.Str(
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    rev_reg_id = fields.Str(
        validate=INDY_REV_REG_ID_VALIDATE,
        allow_none=True,
        metadata={
            "description": "Revocation registry identifier",
            "example": INDY_REV_REG_ID_EXAMPLE,
        },
    )
    cred_rev_id = fields.Str(
        validate=INDY_CRED_REV_ID_VALIDATE,
        allow_none=True,
        metadata={
            "description": "Credential revocation identifier",
            "example": INDY_CRED_REV_ID_EXAMPLE,
        },
    )


class IndyCredPrecisSchema(OpenAPISchema):
    """Schema for precis that indy credential search returns (and aca-py augments)."""

    cred_info = fields.Nested(
        IndyCredInfoSchema(),
        metadata={"description": "Credential info"},
        required=True,
    )
    interval = fields.Nested(
        IndyNonRevocationIntervalSchema(),
        metadata={"description": "Non-revocation interval from presentation request"},
    )
    presentation_referents = fields.List(
        fields.Str(
            required=True,
            metadata={"description": "presentation referent", "example": "1_age_uuid"},
        )
    )



from typing import Coroutine, Union

from ....connections.models.conn_record import ConnRecord
from ....core.error import BaseError
from .messages.presentation_problem_report import (
    PresentationProblemReport,
    ProblemReportReason,
)
from .models.presentation_exchange import V10PresentationExchange


def problem_report_for_record(
    record: Union[ConnRecord, V10PresentationExchange],
    desc_en: str,
) -> PresentationProblemReport:
    """Create problem report for record.

    Args:
        record: connection or exchange record
        desc_en: description text to include in problem report

    """
    result = PresentationProblemReport(
        description={
            "en": desc_en,
            "code": ProblemReportReason.ABANDONED.value,
        },
    )
    if record:
        thid = getattr(record, "thread_id", None)
        if thid:
            result.assign_thread_id(thid)

    return result


async def report_problem(
    err: BaseError,
    desc_en: str,
    http_error_class,
    record: Union[ConnRecord, V10PresentationExchange],
    outbound_handler: Coroutine,
):
    """Send problem report response and raise corresponding HTTP error.

    Args:
        err: error for internal diagnostics
        desc_en: description text to include in problem report (response)
        http_error_class: HTTP error to raise
        record: record to cite by thread in problem report
        outbound_handler: outbound message handler

    """
    if record:
        await outbound_handler(
            problem_report_for_record(record, desc_en),
            connection_id=record.connection_id,
        )

    raise http_error_class(reason=err.roll_up) from err



"""Classes to manage connections."""

import asyncio
import logging
import re
from typing import List, Mapping, NamedTuple, Optional, Sequence, Text, Union

from uuid_utils import uuid4

from ....connections.base_manager import BaseConnectionManager
from ....connections.models.conn_record import ConnRecord
from ....core.error import BaseError
from ....core.event_bus import EventBus
from ....core.oob_processor import OobMessageProcessor
from ....core.profile import Profile
from ....did.did_key import DIDKey
from ....messaging.decorators.attach_decorator import AttachDecorator
from ....messaging.decorators.service_decorator import ServiceDecorator
from ....messaging.responder import BaseResponder
from ....messaging.valid import IndyDID
from ....storage.error import StorageNotFoundError
from ....transport.inbound.receipt import MessageReceipt
from ....wallet.base import BaseWallet
from ....wallet.did_info import INVITATION_REUSE_KEY, DIDInfo
from ....wallet.did_method import PEER2, PEER4
from ....wallet.error import WalletNotFoundError
from ....wallet.key_type import ED25519
from ...coordinate_mediation.v1_0.models.mediation_record import MediationRecord
from ...coordinate_mediation.v1_0.route_manager import RouteManager
from ...didcomm_prefix import DIDCommPrefix
from ...didexchange.v1_0.manager import DIDXManager
from ...issue_credential.v1_0.models.credential_exchange import V10CredentialExchange
from ...issue_credential.v2_0.models.cred_ex_record import V20CredExRecord
from ...present_proof.v1_0.models.presentation_exchange import V10PresentationExchange
from ...present_proof.v2_0.models.pres_exchange import V20PresExRecord
from .message_types import DEFAULT_VERSION
from .messages.invitation import HSProto, InvitationMessage
from .messages.problem_report import OOBProblemReport
from .messages.reuse import HandshakeReuse
from .messages.reuse_accept import HandshakeReuseAccept
from .messages.service import Service
from .messages.service import Service as ServiceMessage
from .models.invitation import InvitationRecord
from .models.oob_record import OobRecord

LOGGER = logging.getLogger(__name__)
REUSE_WEBHOOK_TOPIC = "acapy::webhook::connection_reuse"
REUSE_ACCEPTED_WEBHOOK_TOPIC = "acapy::webhook::connection_reuse_accepted"


class OutOfBandManagerError(BaseError):
    """Out of band error."""


class OutOfBandManagerNotImplementedError(BaseError):
    """Out of band error for unimplemented functionality."""


class InvitationCreator:
    """Class for creating an out of band invitation."""

    class CreateResult(NamedTuple):
        """Result from creating an invitation."""

        invitation_url: str
        invitation: InvitationMessage
        our_recipient_key: str
        connection: Optional[ConnRecord]
        service: Optional[ServiceDecorator]

    def __init__(
        self,
        profile: Profile,
        route_manager: RouteManager,
        oob: "OutOfBandManager",
        my_label: Optional[str] = None,
        my_endpoint: Optional[str] = None,
        auto_accept: Optional[bool] = None,
        public: bool = False,
        use_did: Optional[str] = None,
        use_did_method: Optional[str] = None,
        hs_protos: Optional[Sequence[HSProto]] = None,
        multi_use: bool = False,
        create_unique_did: bool = False,
        alias: Optional[str] = None,
        attachments: Optional[Sequence[Mapping]] = None,
        metadata: Optional[dict] = None,
        mediation_id: Optional[str] = None,
        service_accept: Optional[Sequence[Text]] = None,
        protocol_version: Optional[Text] = None,
        goal_code: Optional[Text] = None,
        goal: Optional[Text] = None,
    ):
        """Initialize the invitation creator."""
        if not (hs_protos or attachments):
            raise OutOfBandManagerError(
                "Invitation must include handshake protocols, "
                "request attachments, or both"
            )

        if not hs_protos and metadata:
            raise OutOfBandManagerError(
                "Cannot store metadata without handshake protocols"
            )

        if attachments and multi_use:
            raise OutOfBandManagerError(
                "Cannot create multi use invitation with attachments"
            )

        if public and use_did:
            raise OutOfBandManagerError("use_did and public are mutually exclusive")

        if public and use_did_method:
            raise OutOfBandManagerError(
                "use_did_method and public are mutually exclusive"
            )

        if use_did and use_did_method:
            raise OutOfBandManagerError(
                "use_did and use_did_method are mutually exclusive"
            )

        if create_unique_did and not use_did_method:
            LOGGER.error(
                "create_unique_did: `%s`, use_did_method: `%s`",
                create_unique_did,
                use_did_method,
            )
            raise OutOfBandManagerError(
                "create_unique_did can only be used with use_did_method"
            )

        if use_did_method and use_did_method not in DIDXManager.SUPPORTED_USE_DID_METHODS:
            raise OutOfBandManagerError(f"Unsupported use_did_method: {use_did_method}")

        self.profile = profile
        self.route_manager = route_manager
        self.oob = oob

        self.msg_id = str(uuid4())
        self.attachments = attachments

        self.handshake_protocols = [
            DIDCommPrefix.qualify_current(hsp.name) for hsp in hs_protos or []
        ] or None

        if not my_endpoint:
            my_endpoint = self.profile.settings.get("default_endpoint")
            assert my_endpoint
        self.my_endpoint = my_endpoint

        self.version = protocol_version or DEFAULT_VERSION

        if not my_label:
            my_label = self.profile.settings.get("default_label")
            assert my_label
        self.my_label = my_label

        self.accept = service_accept if protocol_version != "1.0" else None
        self.invitation_mode = (
            ConnRecord.INVITATION_MODE_MULTI
            if multi_use
            else ConnRecord.INVITATION_MODE_ONCE
        )
        self.alias = alias

        auto_accept = bool(
            auto_accept
            or (
                auto_accept is None
                and self.profile.settings.get("debug.auto_accept_requests")
            )
        )
        self.auto_accept = (
            ConnRecord.ACCEPT_AUTO if auto_accept else ConnRecord.ACCEPT_MANUAL
        )
        self.goal = goal
        self.goal_code = goal_code
        self.public = public
        self.use_did = use_did
        self.use_did_method = use_did_method
        self.multi_use = multi_use
        self.create_unique_did = create_unique_did
        self.image_url = self.profile.context.settings.get("image_url")

        self.mediation_id = mediation_id
        self.metadata = metadata

    async def create_attachment(self, attachment: Mapping, pthid: str) -> AttachDecorator:
        """Create attachment for OOB invitation."""
        a_type = attachment.get("type")
        a_id = attachment.get("id")

        if not a_type or not a_id:
            raise OutOfBandManagerError("Attachment must include type and id")

        async with self.profile.session() as session:
            if a_type == "credential-offer":
                try:
                    cred_ex_rec = await V10CredentialExchange.retrieve_by_id(
                        session,
                        a_id,
                    )
                    message = cred_ex_rec.credential_offer_dict

                except StorageNotFoundError:
                    cred_ex_rec = await V20CredExRecord.retrieve_by_id(
                        session,
                        a_id,
                    )
                    message = cred_ex_rec.cred_offer
            elif a_type == "present-proof":
                try:
                    pres_ex_rec = await V10PresentationExchange.retrieve_by_id(
                        session,
                        a_id,
                    )
                    message = pres_ex_rec.presentation_request_dict
                except StorageNotFoundError:
                    pres_ex_rec = await V20PresExRecord.retrieve_by_id(
                        session,
                        a_id,
                    )
                    message = pres_ex_rec.pres_request
            else:
                raise OutOfBandManagerError(f"Unknown attachment type: {a_type}")

        message.assign_thread_id(pthid=pthid)
        return InvitationMessage.wrap_message(message.serialize())

    async def create_attachments(
        self,
        invitation_msg_id: str,
        attachments: Optional[Sequence[Mapping]] = None,
    ) -> List[AttachDecorator]:
        """Create attachments for OOB invitation."""
        return [
            await self.create_attachment(attachment, invitation_msg_id)
            for attachment in attachments or []
        ]

    async def create(self) -> InvitationRecord:
        """Create the invitation, returning the result as an InvitationRecord."""
        attachments = await self.create_attachments(self.msg_id, self.attachments)
        mediation_record = await self.oob._route_manager.mediation_record_if_id(
            self.profile, self.mediation_id, or_default=True
        )

        if self.public:
            result = await self.handle_public(attachments, mediation_record)
        elif self.use_did:
            result = await self.handle_use_did(attachments, mediation_record)
        elif self.use_did_method:
            result = await self.handle_use_did_method(attachments, mediation_record)
        else:
            result = await self.handle_legacy_invite_key(attachments, mediation_record)

        oob_record = OobRecord(
            role=OobRecord.ROLE_SENDER,
            state=OobRecord.STATE_AWAIT_RESPONSE,
            connection_id=(
                result.connection.connection_id if result.connection else None
            ),
            invi_msg_id=self.msg_id,
            invitation=result.invitation,
            our_recipient_key=result.our_recipient_key,
            our_service=result.service,
            multi_use=self.multi_use,
        )

        async with self.profile.session() as session:
            await oob_record.save(session, reason="Created new oob invitation")

        return InvitationRecord(
            oob_id=oob_record.oob_id,
            state=InvitationRecord.STATE_INITIAL,
            invi_msg_id=self.msg_id,
            invitation=result.invitation,
            invitation_url=result.invitation_url,
        )

    async def handle_handshake_protos(
        self,
        invitation_key: str,
        msg: InvitationMessage,
        mediation_record: Optional[MediationRecord],
    ) -> ConnRecord:
        """Handle handshake protocol options, creating a ConnRecord.

        When handshake protocols are included in the create request, that means
        we intend to create a connection for the invitation. When absent,
        no connection is created, representing a connectionless exchange.
        """
        assert self.handshake_protocols

        if len(self.handshake_protocols) == 1:
            connection_protocol = DIDCommPrefix.unqualify(self.handshake_protocols[0])
        else:
            # We don't know which protocol will be used until the request is received
            connection_protocol = None

        conn_rec = ConnRecord(
            invitation_key=invitation_key,
            invitation_msg_id=self.msg_id,
            invitation_mode=self.invitation_mode,
            their_role=ConnRecord.Role.REQUESTER.rfc23,
            state=ConnRecord.State.INVITATION.rfc23,
            accept=self.auto_accept,
            alias=self.alias,
            connection_protocol=connection_protocol,
        )

        LOGGER.debug("Creating connection record for invitation %s", self.msg_id)
        async with self.profile.transaction() as session:
            await conn_rec.save(session, reason="Created new invitation")
            await conn_rec.attach_invitation(session, msg)

            if self.metadata:
                LOGGER.debug("Setting metadata for connection %s", conn_rec.connection_id)
                for key, value in self.metadata.items():
                    await conn_rec.metadata_set(session, key, value)

            await session.commit()

        LOGGER.debug("Routing invitation %s", conn_rec.connection_id)
        await self.route_manager.route_invitation(
            self.profile, conn_rec, mediation_record
        )

        return conn_rec

    def did_key_to_key(self, did_key: str) -> str:
        """Convert a DID key to a key."""
        if did_key.startswith("did:key:"):
            return DIDKey.from_did(did_key).public_key_b58
        return did_key

    def did_keys_to_keys(self, did_keys: Sequence[str]) -> List[str]:
        """Convert DID keys to keys."""
        return [self.did_key_to_key(did_key) for did_key in did_keys]

    async def handle_did(
        self,
        did_info: DIDInfo,
        attachments: Sequence[AttachDecorator],
        mediation_record: Optional[MediationRecord],
    ) -> CreateResult:
        """Handle use_did invitation creation."""
        LOGGER.debug("Handling invitation using DID %s", did_info.did)
        invi_msg = InvitationMessage(
            _id=self.msg_id,
            label=self.my_label,
            handshake_protocols=self.handshake_protocols,
            requests_attach=attachments or None,
            services=[did_info.did],
            accept=self.accept,
            version=self.version,
            image_url=self.image_url,
        )
        endpoint, recipient_keys, routing_keys = await self.oob.resolve_invitation(
            did_info.did
        )
        invi_url = invi_msg.to_url(endpoint)

        if self.handshake_protocols:
            LOGGER.debug(
                "Handshake protocols given: %s. Creating connection",
                self.handshake_protocols,
            )
            conn_rec = await self.handle_handshake_protos(
                did_info.verkey, invi_msg, mediation_record
            )
            our_service = None
        else:
            conn_rec = None
            LOGGER.debug("No handshake protocols. Routing verkey %s", did_info.verkey)
            await self.route_manager.route_verkey(
                self.profile, did_info.verkey, mediation_record
            )
            our_service = ServiceDecorator(
                recipient_keys=self.did_keys_to_keys(recipient_keys),
                endpoint=self.my_endpoint,
                routing_keys=self.did_keys_to_keys(routing_keys),
            )

        return self.CreateResult(
            invitation_url=invi_url,
            invitation=invi_msg,
            our_recipient_key=did_info.verkey,
            connection=conn_rec,
            service=our_service,
        )

    async def handle_public(
        self,
        attachments: Sequence[AttachDecorator],
        mediation_record: Optional[MediationRecord] = None,
    ) -> CreateResult:
        """Handle public invitation creation."""
        assert self.public
        if not self.profile.settings.get("public_invites"):
            raise OutOfBandManagerError("Public invitations are not enabled")

        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            public_did = await wallet.get_public_did()

        if not public_did:
            raise OutOfBandManagerError(
                "Cannot create public invitation with no public DID"
            )

        LOGGER.debug("Public DID found: %s", public_did.did)
        if bool(IndyDID.PATTERN.match(public_did.did)):
            public_did = DIDInfo(
                did=f"did:sov:{public_did.did}",
                verkey=public_did.verkey,
                metadata=public_did.metadata,
                method=public_did.method,
                key_type=public_did.key_type,
            )

        return await self.handle_did(public_did, attachments, mediation_record)

    async def handle_use_did(
        self,
        attachments: Sequence[AttachDecorator],
        mediation_record: Optional[MediationRecord],
    ) -> CreateResult:
        """Handle use_did invitation creation."""
        assert self.use_did
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            try:
                did_info = await wallet.get_local_did(self.use_did)
            except WalletNotFoundError:
                raise OutOfBandManagerError(
                    f"Cannot find DID for invitation reuse: {self.use_did}"
                )
        return await self.handle_did(did_info, attachments, mediation_record)

    async def handle_use_did_method(
        self,
        attachments: Sequence[AttachDecorator],
        mediation_record: Optional[MediationRecord],
    ) -> CreateResult:
        """Create an invitation using a DID method, optionally reusing one."""
        assert self.use_did_method
        mediation_records = [mediation_record] if mediation_record else []

        if self.my_endpoint:
            my_endpoints = [self.my_endpoint]
        else:
            my_endpoints = []
            default_endpoint = self.profile.settings.get("default_endpoint")
            if default_endpoint:
                my_endpoints.append(default_endpoint)
            my_endpoints.extend(self.profile.settings.get("additional_endpoints", []))

        did_peer_4 = self.use_did_method == "did:peer:4"

        my_info = None
        if not self.create_unique_did:
            # check wallet to see if there is an existing "invitation" DID available
            did_method = PEER4 if did_peer_4 else PEER2
            my_info = await self.oob.fetch_invitation_reuse_did(did_method)
            if not my_info:
                LOGGER.warning("No invitation DID found, creating new DID")

        if not my_info:
            did_metadata = (
                {INVITATION_REUSE_KEY: "true"} if not self.create_unique_did else {}
            )
            if did_peer_4:
                my_info = await self.oob.create_did_peer_4(
                    my_endpoints, mediation_records, did_metadata
                )
            else:
                my_info = await self.oob.create_did_peer_2(
                    my_endpoints, mediation_records, did_metadata
                )

        return await self.handle_did(my_info, attachments, mediation_record)

    async def handle_legacy_invite_key(
        self,
        attachments: Sequence[AttachDecorator],
        mediation_record: Optional[MediationRecord],
    ) -> CreateResult:
        """Create an invitation using legacy bare public key and inline service."""
        LOGGER.debug("Handling legacy invitation")
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            connection_key = await wallet.create_signing_key(ED25519)

        routing_keys, routing_endpoint = await self.route_manager.routing_info(
            self.profile, mediation_record
        )
        routing_keys = [
            (
                key
                if len(key.split(":")) == 3
                else DIDKey.from_public_key_b58(key, ED25519).key_id
            )
            for key in routing_keys or []
        ]
        recipient_keys = [
            DIDKey.from_public_key_b58(connection_key.verkey, ED25519).key_id
        ]

        my_endpoint = routing_endpoint or self.my_endpoint

        invi_msg = InvitationMessage(
            _id=self.msg_id,
            label=self.my_label,
            handshake_protocols=self.handshake_protocols,
            requests_attach=attachments,
            accept=self.accept,
            image_url=self.image_url,
            version=self.version,
            services=[
                ServiceMessage(
                    _id="#inline",
                    _type="did-communication",
                    recipient_keys=recipient_keys,
                    service_endpoint=my_endpoint,
                    routing_keys=routing_keys,
                )
            ],
            goal=self.goal,
            goal_code=self.goal_code,
        )

        if self.handshake_protocols:
            LOGGER.debug("Handshake protocols given: %s", self.handshake_protocols)
            conn_rec = await self.handle_handshake_protos(
                connection_key.verkey, invi_msg, mediation_record
            )
            our_service = None
        else:
            LOGGER.debug(
                "No handshake protocols. Routing verkey %s", connection_key.verkey
            )
            await self.route_manager.route_verkey(
                self.profile, connection_key.verkey, mediation_record
            )
            conn_rec = None
            our_service = ServiceDecorator(
                recipient_keys=self.did_keys_to_keys(recipient_keys),
                endpoint=my_endpoint,
                routing_keys=self.did_keys_to_keys(routing_keys),
            )

        return self.CreateResult(
            invitation_url=invi_msg.to_url(),
            invitation=invi_msg,
            our_recipient_key=connection_key.verkey,
            connection=conn_rec,
            service=our_service,
        )


class OutOfBandManager(BaseConnectionManager):
    """Class for managing out of band messages."""

    def __init__(self, profile: Profile):
        """Initialize a OutOfBandManager.

        Args:
            profile: The profile for this out of band manager
        """
        self._profile = profile
        super().__init__(self._profile)

    @property
    def profile(self) -> Profile:
        """Accessor for the current profile.

        Returns:
            The profile for this connection manager

        """
        return self._profile

    async def create_invitation(
        self,
        my_label: Optional[str] = None,
        my_endpoint: Optional[str] = None,
        auto_accept: Optional[bool] = None,
        public: bool = False,
        use_did: Optional[str] = None,
        use_did_method: Optional[str] = None,
        hs_protos: Optional[Sequence[HSProto]] = None,
        multi_use: bool = False,
        create_unique_did: bool = False,
        alias: Optional[str] = None,
        attachments: Optional[Sequence[Mapping]] = None,
        metadata: Optional[dict] = None,
        mediation_id: Optional[str] = None,
        service_accept: Optional[Sequence[Text]] = None,
        protocol_version: Optional[Text] = None,
        goal_code: Optional[Text] = None,
        goal: Optional[Text] = None,
    ) -> InvitationRecord:
        """Generate new connection invitation.

        This method generates a new connection invitation, which represents an out-of-band
        communication channel. In the future and in practice, these sort of invitations
        will be received over any number of channels such as SMS, Email, QR Code, NFC,
        etc.

        Args:
            my_label (Optional[str]): Label for this connection.
            my_endpoint (Optional[str]): Endpoint where the other party can reach me.
            auto_accept (Optional[bool]): Auto-accept a corresponding connection request
                (None to use config).
            public (bool): Set to True to create an invitation from the public DID.
            use_did (Optional[str]): DID to use for the invitation.
            use_did_method (Optional[str]): DID method to use for the invitation.
            hs_protos (Optional[Sequence[HSProto]]): List of handshake protocols to
                include.
            multi_use (bool): Set to True to create an invitation for multiple-use
                connection.
            create_unique_did (bool): Set to True to create a unique DID for the
                invitation.
            alias (Optional[str]): Optional alias to apply to the connection for later
                use.
            attachments (Optional[Sequence[Mapping]]): List of attachments in the form of
                {"id": ..., "type": ...}.
            metadata (Optional[dict]): Additional metadata for the invitation.
            mediation_id (Optional[str]): Mediation ID for the invitation.
            service_accept (Optional[Sequence[Text]]): Optional list of mime types in the
                order of preference of the sender that the receiver can use in responding
                to the message.
            protocol_version (Optional[Text]): OOB protocol version [1.0, 1.1].
            goal_code (Optional[Text]): Optional self-attested code for receiver logic.
            goal (Optional[Text]): Optional self-attested string for receiver logic.

        Returns:
            InvitationRecord: The generated invitation record.

        """
        creator = InvitationCreator(
            self.profile,
            self._route_manager,
            self,
            my_label,
            my_endpoint,
            auto_accept,
            public,
            use_did,
            use_did_method,
            hs_protos,
            multi_use,
            create_unique_did,
            alias,
            attachments,
            metadata,
            mediation_id,
            service_accept,
            protocol_version,
            goal_code,
            goal,
        )
        return await creator.create()

    async def receive_invitation(
        self,
        invitation: InvitationMessage,
        use_existing_connection: bool = True,
        auto_accept: Optional[bool] = None,
        alias: Optional[str] = None,
        mediation_id: Optional[str] = None,
    ) -> OobRecord:
        """Receive an out of band invitation message.

        Args:
            invitation: invitation message
            use_existing_connection: whether to use existing connection if possible
            auto_accept: whether to accept the invitation automatically
            alias: Alias for connection record
            mediation_id: mediation identifier

        Returns:
            ConnRecord, serialized

        """
        if mediation_id:
            try:
                LOGGER.debug("Getting mediation record for %s", mediation_id)
                await self._route_manager.mediation_record_if_id(
                    self.profile, mediation_id
                )
            except StorageNotFoundError:
                mediation_id = None

        # There must be exactly 1 service entry
        if len(invitation.services) != 1:
            raise OutOfBandManagerError("service array must have exactly one element")

        if not (invitation.requests_attach or invitation.handshake_protocols):
            raise OutOfBandManagerError(
                "Invitation must specify handshake_protocols, requests_attach, or both"
            )

        # Get the single service item
        oob_service_item = invitation.services[0]

        # service_accept
        service_accept = invitation.accept

        # Get the DID public did, if any (might also be a did:peer)
        public_did = None
        if isinstance(oob_service_item, str):
            if bool(IndyDID.PATTERN.match(oob_service_item)):
                public_did = oob_service_item.split(":")[-1]
            else:
                public_did = oob_service_item

        conn_rec = None

        # Find existing connection - only if started by an invitation with Public DID
        # (or did:peer) and use_existing_connection is true
        if (
            public_did is not None and use_existing_connection
        ):  # invite has public DID: seek existing connection
            if public_did.startswith("did:peer:4"):
                search_public_did = self.long_did_peer_to_short(public_did)
            else:
                search_public_did = public_did

            LOGGER.debug(
                "Trying to find existing connection for oob invitation with did %s",
                search_public_did,
            )

            async with self._profile.session() as session:
                conn_rec = await ConnRecord.find_existing_connection(
                    session=session, their_public_did=search_public_did
                )

        oob_record = OobRecord(
            role=OobRecord.ROLE_RECEIVER,
            invi_msg_id=invitation._id,
            invitation=invitation,
            state=OobRecord.STATE_INITIAL,
            connection_id=conn_rec.connection_id if conn_rec else None,
        )

        # Try to reuse the connection. If not accepted sets the conn_rec to None
        if conn_rec and not invitation.requests_attach:
            oob_record = await self._handle_handshake_reuse(
                oob_record, conn_rec, invitation._version
            )

            LOGGER.warning(
                "Connection reuse request finished with state %s", oob_record.state
            )

            if oob_record.state == OobRecord.STATE_ACCEPTED:
                return oob_record
            else:
                # Set connection record to None if not accepted
                # Will make new connection
                conn_rec = None

        # Try to create a connection. Either if the reuse failed or we didn't have a
        # connection yet. Throws an error if connection could not be created
        if not conn_rec and invitation.handshake_protocols:
            oob_record = await self._perform_handshake(
                oob_record=oob_record,
                alias=alias,
                auto_accept=auto_accept,
                mediation_id=mediation_id,
                service_accept=service_accept,
            )
            LOGGER.debug(
                "Performed handshake with connection %s", oob_record.connection_id
            )
            # re-fetch connection record
            async with self.profile.session() as session:
                conn_rec = await ConnRecord.retrieve_by_id(
                    session, oob_record.connection_id
                )

        # If a connection record is associated with the oob record we can remove it now as
        # we can leverage the connection for all exchanges. Otherwise we need to keep it
        # around for the connectionless exchange
        if conn_rec:
            oob_record.state = OobRecord.STATE_DONE
            async with self.profile.session() as session:
                await oob_record.emit_event(session)
                await oob_record.delete_record(session)
        else:
            oob_record.state = OobRecord.STATE_PREPARE_RESPONSE
            async with self.profile.session() as session:
                await oob_record.save(session)

        # Handle any attachments
        if invitation.requests_attach:
            LOGGER.debug(
                "Process attached messages for oob exchange %s (connection_id %s)",
                oob_record.oob_id,
                oob_record.connection_id,
            )

            # FIXME: this should ideally be handled using an event handler. Once the
            # connection is ready we start processing the attached messages.
            # For now we use the timeout method
            if (
                conn_rec
                and not conn_rec.is_ready
                and not await self._wait_for_conn_rec_active(conn_rec.connection_id)
            ):
                raise OutOfBandManagerError(
                    "Connection not ready to process attach message "
                    f"for connection_id: {oob_record.connection_id} and "
                    f"invitation_msg_id {invitation._id}",
                )

            if not conn_rec:
                # Create and store new key for connectionless exchange
                async with self.profile.session() as session:
                    wallet = session.inject(BaseWallet)
                    connection_key = await wallet.create_signing_key(ED25519)
                    oob_record.our_recipient_key = connection_key.verkey
                    oob_record.our_service = ServiceDecorator(
                        recipient_keys=[connection_key.verkey],
                        endpoint=self.profile.settings.get("default_endpoint"),
                        routing_keys=[],
                    ).serialize()

                    # Need to make sure the created key is routed by the base wallet
                    await self._route_manager.route_verkey(
                        self.profile, connection_key.verkey
                    )
                    await oob_record.save(session)

            await self._process_request_attach(oob_record)

        return oob_record

    async def _process_request_attach(self, oob_record: OobRecord):
        invitation = oob_record.invitation

        message_processor = self.profile.inject(OobMessageProcessor)
        messages = [attachment.content for attachment in invitation.requests_attach]

        their_service = None
        if not oob_record.connection_id:
            service = oob_record.invitation.services[0]
            their_service = await self._service_decorator_from_service(service)
            if their_service:
                LOGGER.debug("Found service for oob record %s", their_service)
            else:
                LOGGER.debug("No service decorator obtained from %s", service)

        await message_processor.handle_message(
            self.profile, messages, oob_record=oob_record, their_service=their_service
        )

    async def _service_decorator_from_service(
        self, service: Union[Service, str]
    ) -> Optional[ServiceDecorator]:
        if isinstance(service, str):
            (
                endpoint,
                recipient_keys,
                routing_keys,
            ) = await self.resolve_invitation(service)

            if not endpoint:
                return None

            return ServiceDecorator(
                endpoint=endpoint,
                recipient_keys=recipient_keys,
                routing_keys=routing_keys,
            )
        elif isinstance(service, Service):
            endpoint = service.service_endpoint

            if not endpoint:
                return None

            recipient_keys = [
                DIDKey.from_did(did_key).public_key_b58
                for did_key in service.recipient_keys
            ]
            routing_keys = [
                DIDKey.from_did(did_key).public_key_b58
                for did_key in service.routing_keys
            ]

            return ServiceDecorator(
                endpoint=endpoint,
                recipient_keys=recipient_keys,
                routing_keys=routing_keys,
            )
        else:
            LOGGER.warning(
                "Unexpected type `%s` passed to `_service_decorator_from_service`",
                type(service),
            )
            return None

    async def _wait_for_reuse_response(self, oob_id: str, timeout: int = 15) -> OobRecord:
        """Wait for reuse response.

        Wait for reuse response message state. Either by receiving a reuse accepted or
        problem report. If no answer is received within the timeout, the state will be
        set to reuse_not_accepted

        Args:
            oob_id: Identifier of the oob record
            timeout: The timeout in seconds to wait for the reuse state [default=15]

        Returns:
            OobRecord: The oob record associated with the provided id.

        """
        OOB_REUSE_RESPONSE_STATE = re.compile(
            "^acapy::record::out_of_band::(reuse-accepted|reuse-not-accepted)$"
        )

        async def _wait_for_state() -> OobRecord:
            event = self.profile.inject(EventBus)
            with event.wait_for_event(
                self.profile,
                OOB_REUSE_RESPONSE_STATE,
                lambda event: event.payload.get("oob_id") == oob_id,
            ) as await_event:
                # After starting the listener first retrieve the record from storage.
                # This rules out the scenario where the record was in the desired state
                # Before starting the event listener
                async with self.profile.session() as session:
                    oob_record = await OobRecord.retrieve_by_id(session, oob_id)

                    if oob_record.state in [
                        OobRecord.STATE_ACCEPTED,
                        OobRecord.STATE_NOT_ACCEPTED,
                    ]:
                        return oob_record

                LOGGER.debug("Wait for oob %s to receive reuse accepted message", oob_id)
                event = await await_event
                LOGGER.debug("Received reuse response message")
                return OobRecord.deserialize(event.payload)

        try:
            oob_record = await asyncio.wait_for(
                _wait_for_state(),
                timeout,
            )

            return oob_record
        except asyncio.TimeoutError:
            async with self.profile.session() as session:
                oob_record = await OobRecord.retrieve_by_id(session, oob_id)
                return oob_record

    async def _wait_for_conn_rec_active(
        self, connection_id: str, timeout: int = 7
    ) -> Optional[ConnRecord]:
        CONNECTION_READY_EVENT = re.compile(
            "^acapy::record::connections::(active|completed|response)$"
        )

        LOGGER.debug("Wait for connection %s to become active", connection_id)

        async def _wait_for_state() -> ConnRecord:
            event = self.profile.inject(EventBus)
            with event.wait_for_event(
                self.profile,
                CONNECTION_READY_EVENT,
                lambda event: event.payload.get("connection_id") == connection_id,
            ) as await_event:
                # After starting the listener first retrieve the record from storage.
                # This rules out the scenario where the record was in the desired state
                # Before starting the event listener
                async with self.profile.session() as session:
                    conn_record = await ConnRecord.retrieve_by_id(session, connection_id)
                    if conn_record.is_ready:
                        return conn_record

                LOGGER.debug("Wait for connection %s to become active", connection_id)
                # Wait for connection record to be in state
                event = await await_event
                return ConnRecord.deserialize(event.payload)

        try:
            return await asyncio.wait_for(
                _wait_for_state(),
                timeout,
            )

        except asyncio.TimeoutError:
            LOGGER.warning("Connection for connection_id %s not ready", connection_id)
            return None

    async def _handle_handshake_reuse(
        self, oob_record: OobRecord, conn_record: ConnRecord, version: str
    ) -> OobRecord:
        # Send handshake reuse
        oob_record = await self._create_handshake_reuse_message(
            oob_record, conn_record, version
        )

        # Wait for the reuse accepted message
        oob_record = await self._wait_for_reuse_response(oob_record.oob_id)
        LOGGER.debug(
            "Oob reuse for oob id %s with connection %s finished with state %s",
            oob_record.oob_id,
            oob_record.connection_id,
            oob_record.state,
        )

        if oob_record.state != OobRecord.STATE_ACCEPTED:
            # Remove associated connection id as reuse has ben denied
            oob_record.connection_id = None
            oob_record.state = OobRecord.STATE_NOT_ACCEPTED

            # OOB_TODO: replace webhook event with new oob webhook event
            # Emit webhook if the reuse was not accepted
            await self.profile.notify(
                REUSE_ACCEPTED_WEBHOOK_TOPIC,
                {
                    "thread_id": oob_record.reuse_msg_id,
                    "connection_id": conn_record.connection_id,
                    "state": "rejected",
                    "comment": (
                        "No HandshakeReuseAccept message received, "
                        f"connection {conn_record.connection_id} ",
                        f"and invitation {oob_record.invitation._id}",
                    ),
                },
            )

            async with self.profile.session() as session:
                await oob_record.save(session)

        return oob_record

    async def _perform_handshake(
        self,
        *,
        oob_record: OobRecord,
        alias: Optional[str] = None,
        auto_accept: Optional[bool] = None,
        mediation_id: Optional[str] = None,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> OobRecord:
        invitation = oob_record.invitation

        supported_handshake_protocols = [
            HSProto.get(DIDCommPrefix.unqualify(proto))
            for proto in invitation.handshake_protocols
        ]

        # Get the single service item
        service = invitation.services[0]
        public_did = None
        if isinstance(service, str):
            # If it's in the did format, we need to convert to a full service block
            # An existing connection can only be reused based on a public DID
            # in an out-of-band message (RFC 0434).
            # OR did:peer:2 or did:peer:4.

            if service.startswith("did:peer"):
                public_did = service
                if public_did.startswith("did:peer:4"):
                    public_did = self.long_did_peer_to_short(public_did)
            else:
                public_did = service.split(":")[-1]

            # TODO: resolve_invitation should resolve key_info objects
            # or something else that includes the key type. We now assume
            # ED25519 keys
            endpoint, recipient_keys, routing_keys = await self.resolve_invitation(
                service,
                service_accept=service_accept,
            )
            service = ServiceMessage.deserialize(
                {
                    "id": "#inline",
                    "type": "did-communication",
                    "recipientKeys": [
                        DIDKey.from_public_key_b58(key, ED25519).key_id
                        for key in recipient_keys
                    ],
                    "routingKeys": [
                        DIDKey.from_public_key_b58(key, ED25519).key_id
                        for key in routing_keys
                    ],
                    "serviceEndpoint": endpoint,
                }
            )

        if public_did:
            LOGGER.debug("Creating connection with public did %s", public_did)
        else:
            LOGGER.debug("Creating connection with service %s", service)

        conn_record = None
        for protocol in supported_handshake_protocols:
            # DIDExchange
            if protocol is HSProto.RFC23 or protocol is HSProto.DIDEX_1_1:
                didx_mgr = DIDXManager(self.profile)
                conn_record = await didx_mgr.receive_invitation(
                    invitation=invitation,
                    their_public_did=public_did,
                    auto_accept=auto_accept,
                    alias=alias,
                    mediation_id=mediation_id,
                    protocol=protocol.name,
                )
                break
        if not conn_record:
            raise OutOfBandManagerError(
                f"Unable to create connection. Could not perform handshake using any of "
                f"the handshake_protocols (supported {supported_handshake_protocols})"
            )

        async with self.profile.session() as session:
            oob_record.connection_id = conn_record.connection_id
            await oob_record.save(session)

        return oob_record

    async def _create_handshake_reuse_message(
        self,
        oob_record: OobRecord,
        conn_record: ConnRecord,
        version: str,
    ) -> OobRecord:
        """Create and Send a Handshake Reuse message under RFC 0434.

        Args:
            oob_record: OOB Record
            conn_record: Connection record associated with the oob record
            version: The version of the OOB protocol

        Returns:
            OobRecord: The oob record with updated state and reuse_msg_id.

        Raises:
            OutOfBandManagerError: If there is an issue creating or
            sending the OOB invitation

        """
        try:
            reuse_msg = HandshakeReuse(version=version)
            reuse_msg.assign_thread_id(thid=reuse_msg._id, pthid=oob_record.invi_msg_id)

            connection_targets = await self.fetch_connection_targets(
                connection=conn_record
            )

            responder = self.profile.inject(BaseResponder)
            await responder.send(
                message=reuse_msg,
                target_list=connection_targets,
            )

            async with self.profile.session() as session:
                oob_record.reuse_msg_id = reuse_msg._id
                oob_record.state = OobRecord.STATE_AWAIT_RESPONSE
                await oob_record.save(session, reason="Storing reuse msg data")

            return oob_record

        except Exception as err:
            raise OutOfBandManagerError(
                f"Error on creating and sending a handshake reuse message: {err}"
            )

    async def delete_stale_connection_by_invitation(self, invi_msg_id: str):
        """Delete unused connections, using existing an active connection instead."""
        tag_filter = {
            "invitation_msg_id": invi_msg_id,
        }
        post_filter = {"invitation_mode": "once", "state": "invitation"}

        async with self.profile.session() as session:
            conn_records = await ConnRecord.query(
                session,
                tag_filter=tag_filter,
                post_filter_positive=post_filter,
            )
            for conn_rec in conn_records:
                await conn_rec.delete_record(session)

    async def fetch_oob_invitation_record_by_id(self, oob_id: str) -> OobRecord:
        """Fetch oob_record associated with an oob_id."""
        async with self.profile.session() as session:
            oob_record = await OobRecord.retrieve_by_id(
                session,
                record_id=oob_id,
            )

        if not oob_record:
            raise StorageNotFoundError(f"No record found with oob_id {oob_id}")

        return oob_record

    async def delete_conn_and_oob_record_invitation(self, invi_msg_id: str):
        """Delete conn_record and oob_record associated with an invi_msg_id."""
        async with self.profile.session() as session:
            conn_records = await ConnRecord.query(
                session,
                tag_filter={
                    "invitation_msg_id": invi_msg_id,
                },
            )
            for conn_rec in conn_records:
                await conn_rec.delete_record(session)
            oob_records = await OobRecord.query(
                session,
                tag_filter={
                    "invi_msg_id": invi_msg_id,
                },
            )
            for oob_rec in oob_records:
                await oob_rec.delete_record(session)

    async def receive_reuse_message(
        self,
        reuse_msg: HandshakeReuse,
        receipt: MessageReceipt,
        conn_rec: ConnRecord,
    ) -> None:
        """Receive and process a HandshakeReuse message under RFC 0434.

        Process a `HandshakeReuse` message by looking up
        the connection records using the MessageReceipt sender DID.

        Args:
            reuse_msg: The `HandshakeReuse` to process
            receipt: The message receipt
            conn_rec: The connection record associated with the message

        Returns:
            None

        Raises:
            OutOfBandManagerError: If the existing connection is not active
            or the connection does not exists

        """
        invi_msg_id = reuse_msg._thread.pthid
        reuse_msg_id = reuse_msg._thread_id

        reuse_accept_msg = HandshakeReuseAccept(version=reuse_msg._version)
        reuse_accept_msg.assign_thread_id(thid=reuse_msg_id, pthid=invi_msg_id)
        connection_targets = await self.fetch_connection_targets(connection=conn_rec)

        responder = self.profile.inject(BaseResponder)

        # Update ConnRecord's invi_msg_id
        async with self._profile.session() as session:
            oob_record = await OobRecord.retrieve_by_tag_filter(
                session,
                {"invi_msg_id": invi_msg_id},
                {"state": OobRecord.STATE_AWAIT_RESPONSE},
            )

            oob_record.state = OobRecord.STATE_DONE
            oob_record.reuse_msg_id = reuse_msg_id
            oob_record.connection_id = conn_rec.connection_id

            # We don't want to store this state. We either remove the record
            # (no multi-use) or we can't update the record (multi-use)
            await oob_record.emit_event(session)

            # If the oob_record is not multi-use we can now remove it
            if not oob_record.multi_use:
                await oob_record.delete_record(session)

            conn_rec.invitation_msg_id = invi_msg_id
            await conn_rec.save(session, reason="Assigning new invitation_msg_id")

        # Delete the ConnRecord created; re-use existing connection
        await self.delete_stale_connection_by_invitation(invi_msg_id)
        # Emit webhook
        await self.profile.notify(
            REUSE_WEBHOOK_TOPIC,
            {
                "thread_id": reuse_msg_id,
                "connection_id": conn_rec.connection_id,
                "comment": (
                    f"Connection {conn_rec.connection_id} is being reused "
                    f"for invitation {invi_msg_id}"
                ),
            },
        )

        await responder.send(
            message=reuse_accept_msg,
            target_list=connection_targets,
        )

    async def receive_reuse_accepted_message(
        self,
        reuse_accepted_msg: HandshakeReuseAccept,
        receipt: MessageReceipt,
        conn_record: ConnRecord,
    ) -> None:
        """Receive and process a HandshakeReuseAccept message under RFC 0434.

        Process a `HandshakeReuseAccept` message by updating the OobRecord
        state to `accepted`.

        Args:
            reuse_accepted_msg: The `HandshakeReuseAccept` to process
            receipt: The message receipt
            conn_record: The connection record associated with the message

        Returns:
            None

        Raises:
            OutOfBandManagerError: if there is an error in processing the
            HandshakeReuseAccept message

        """
        invi_msg_id = reuse_accepted_msg._thread.pthid
        thread_reuse_msg_id = reuse_accepted_msg._thread.thid

        try:
            async with self.profile.session() as session:
                oob_record = await OobRecord.retrieve_by_tag_filter(
                    session,
                    {"invi_msg_id": invi_msg_id, "reuse_msg_id": thread_reuse_msg_id},
                )

                oob_record.state = OobRecord.STATE_ACCEPTED
                oob_record.connection_id = conn_record.connection_id

                # We can now remove the oob_record
                await oob_record.emit_event(session)
                await oob_record.delete_record(session)

                conn_record.invitation_msg_id = invi_msg_id
                await conn_record.save(session, reason="Assigning new invitation_msg_id")
            # Emit webhook
            await self.profile.notify(
                REUSE_ACCEPTED_WEBHOOK_TOPIC,
                {
                    "thread_id": thread_reuse_msg_id,
                    "connection_id": conn_record.connection_id,
                    "state": "accepted",
                    "comment": (
                        f"Connection {conn_record.connection_id} is being reused "
                        f"for invitation {invi_msg_id}"
                    ),
                },
            )
        except Exception as e:
            # Emit webhook
            await self.profile.notify(
                REUSE_ACCEPTED_WEBHOOK_TOPIC,
                {
                    "thread_id": thread_reuse_msg_id,
                    "connection_id": conn_record.connection_id,
                    "state": "rejected",
                    "comment": (
                        "Unable to process HandshakeReuseAccept message, "
                        f"connection {conn_record.connection_id} "
                        f"and invitation {invi_msg_id}"
                    ),
                },
            )
            raise OutOfBandManagerError(
                (
                    "Error processing reuse accepted message "
                    f"for OOB invitation {invi_msg_id}, {e}"
                )
            )

    async def receive_problem_report(
        self,
        problem_report: OOBProblemReport,
        receipt: MessageReceipt,
        conn_record: ConnRecord,
    ) -> None:
        """Receive and process a ProblemReport message from the inviter to invitee.

        Process a `ProblemReport` message by updating the OobRecord
        state to `not_accepted`.

        Args:
            problem_report: The `OOBProblemReport` to process
            receipt: The message receipt
            conn_record: The connection record associated with the OOB record

        Returns:
            None

        Raises:
            OutOfBandManagerError: if there is an error in processing the
            HandshakeReuseAccept message

        """
        invi_msg_id = problem_report._thread.pthid
        thread_reuse_msg_id = problem_report._thread.thid
        try:
            async with self.profile.session() as session:
                oob_record = await OobRecord.retrieve_by_tag_filter(
                    session,
                    {"invi_msg_id": invi_msg_id, "reuse_msg_id": thread_reuse_msg_id},
                )
                oob_record.state = OobRecord.STATE_NOT_ACCEPTED
                await oob_record.save(session)
        except Exception as e:
            raise OutOfBandManagerError(
                (
                    "Error processing problem report message "
                    f"for OOB invitation {invi_msg_id}, {e}"
                )
            )



"""Inner structure of keylist message. Represents a single item in keylist.keys."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from ......messaging.models.base import BaseModel, BaseModelSchema
from ......messaging.valid import DID_KEY_EXAMPLE, DID_KEY_VALIDATE
from ...normalization import normalize_from_public_key


class KeylistKey(BaseModel):
    """Inner structure of Keylist keys attribute."""

    class Meta:
        """KeylistKey metadata."""

        schema_class = "KeylistKeySchema"

    def __init__(
        self,
        *,
        recipient_key: Optional[str] = None,
        action: Optional[str] = None,
        result: Optional[str] = None,
        **kwargs,
    ):
        """Initialize a KeylistKey instance.

        Args:
            recipient_key: The recipient verkey of the route
            action: The requested action to perform
            result: The result of the requested action
            kwargs: Additional key word arguments for BaseRecord

        """
        super().__init__(**kwargs)
        self.recipient_key = normalize_from_public_key(recipient_key)


class KeylistKeySchema(BaseModelSchema):
    """KeylistKey schema."""

    class Meta:
        """KeylistKeySchema metadata."""

        model_class = KeylistKey
        unknown = EXCLUDE

    recipient_key = fields.Str(
        required=True, validate=DID_KEY_VALIDATE, metadata={"example": DID_KEY_EXAMPLE}
    )



from unittest import IsolatedAsyncioTestCase

from ....did.did_key import DIDKey
from ....protocols.didexchange.v1_0.messages.request import DIDXRequest
from ....protocols.out_of_band.v1_0.messages.invitation import InvitationMessage
from ....protocols.out_of_band.v1_0.messages.service import Service
from ....storage.base import BaseStorage
from ....storage.error import StorageNotFoundError
from ....utils.testing import create_test_profile
from ....wallet.key_type import ED25519
from ..conn_record import ConnRecord


class TestConnRecord(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()

        self.test_seed = "testseed000000000000000000000001"
        self.test_did = "55GkHamhTU1ZbTbV2ab9DE"
        self.test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
        self.test_endpoint = "http://localhost"

        self.test_target_did = "GbuDUYXaUZRfHD2jeDuQuP"
        self.test_target_verkey = "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"

        self.test_did_peer_4_a = "did:peer:4zQmV3Hf1TT4Xn73MBVf2NAWdMwrzUabpEvwtV3RoZc17Vxr:z2pfttj3xn6tJ7wpHV9ZSwpQVMNtHC7EtM36r1mC5fDpZ25882Yitk21QPbqzuefKPrbFsexWmQtE78vYWweckXtKeu5BhuFDvCjMUf8SC5z7cMPvp8SCdcbqWnHxygjBH9zAAs9myGRnZYXuAkq6CfBdn6ZiNmdRf65TdVfE3cYfS4jNzVZDs1abwytn4jdFJ2fwVegPB3vLY8XxeUEx12a4rtjkqMhs6zBQbJvc4PVUM9rvMbPM2QeXDy7ovkkHaKLUbNUxjQrcQeiR8MTLe1iaVtUv6RpBf4z7ioqfa4VDRmAZT7isVM3NvENUceeUfDZoFbM8PZqGkCbFvfoKiK3SrmTsvPtpXaBAfR4z7w18cFjsvvLBNMZbPnARn4oZijCkYwgaNmAUthgDP4XBFetdUo8728w25FUwTWjAPc1BdSSWPWMRKwCqyAP1Q1hM8dU6otT27MQaQ1rozKncn3U48CXEi2Ef26EDBrSozEWR273ancFojNXBbZVghZG5b6xdypjQir9PgTF94dsygtu47hNxQweVKLUM1p9umqHLhjvLhpS1aGQkGZNnKUHjLDHdToigo15F7TAf8RfMaducHBThFzEp9TUJmiZFTUYQ1uaBgSPMSaWnvTfUoFmLoGbdrWj1vVEsRrARq37u1SJGLqBx7FM2SUd8nxPsChP5jY8ka8F8r7j8qZLHZqvUXbynPUsViwwdFFk8SCsBWfiQgvq7sRiTdLnYv3H5DSwA1uW2GNYXGgkT9aJza4Sk1gvag5iAbQZgxbU594enjVSTjiWsFw2oYQ75JJwiSEgsP2rhpGsNhXxfECNLUtb7FQbDQPtUvLHCJATf7QXJEoWjpfAywmB6NyQcXfskco6FKJNNHeZBnST6U1meH98Ku66vha1k8hAc72iBhXQBnWUjaGRyzELsh2LkBH2UNwW9TuFhxz3SKtL5pGShVQ5XGQhmdrkWP68d6h7c1JqsfogcDBnmWS4VSbJwgtsPNTSsTHGX8hpGvg"
        self.test_did_peer_4_short_a = (
            "did:peer:4zQmV3Hf1TT4Xn73MBVf2NAWdMwrzUabpEvwtV3RoZc17Vxr"
        )
        self.test_did_peer_4_b = "did:peer:4zQmQ4dEtoGcivpiH6gtWwhWJY2ENVWuZifb62uzR76HGPPw:z7p4QX8zEXt2sMjv1Tqq8Lv8Nx8oGo2uRczBe21vyfMhQzsWDnwGmjriYfUX75WDq622czcdHjWGhh2VTbzKhLXUjY8Ma7g64dKAVcy8SaxN5QVdjwpXgD7htKCgCjah8jHEzyBZFrtdfTHiVXfSUz1BiURQf1Z3NfxW5cWYsvDJVvQzVmdHb8ekzCnvxCqL2UV1v9SBb1DsU66N3PCp9HVpSrqUJQyFU2Ddc8bb6u8SJfBU1nyCkNMgfA1zAyKnSBrzZWyyNzAm9oBV36qjC1Qjfcpq4FBnGr7foh5sLXppBwu2ES8U2nxdGrQzAbN47DKBoKJqPVxNh5tTuBdYjDGt7PcvZQjHQGNXXuhJctM5besZci2saGefCHzoZ87vSsFuKq6oXEsW512eadiNZWjHSdG9J4ToMEMK9WT66vGGLFdZszB3xhdFqEDnAMcpnoFUL5WN243aH6492jPC2Zjdi1BvHC1J8bUuvyihAKXF3WmFz7gJWmh6MrTEWNqb17K6tqbyXjFmfnS2RbAi8xBFj3sSsXkSs6TRTXAZD9DenYaQq4RMa2Kqh6VKGvkXAjVHKcPh9Ncpt6rU9ZYttNHbDJFgahwB8KisVBK8FBpG"
        self.test_did_peer_4_short_b = (
            "did:peer:4zQmQ4dEtoGcivpiH6gtWwhWJY2ENVWuZifb62uzR76HGPPw"
        )
        self.test_did_key = DIDKey.from_public_key_b58(self.test_verkey, ED25519)

        self.test_conn_record = ConnRecord(
            my_did=self.test_did,
            their_did=self.test_target_did,
            their_role=ConnRecord.Role.REQUESTER.rfc23,
            state=ConnRecord.State.COMPLETED.rfc23,
        )
        assert self.test_conn_record.their_role == ConnRecord.Role.REQUESTER.rfc160
        assert self.test_conn_record.state == ConnRecord.State.COMPLETED.rfc160
        assert self.test_conn_record.rfc23_state == ConnRecord.State.COMPLETED.rfc23

    async def test_get_enums(self):
        assert ConnRecord.Role.get("Larry") is None
        assert ConnRecord.State.get("a suffusion of yellow") is None

        assert ConnRecord.Role.get(ConnRecord.Role.REQUESTER) is ConnRecord.Role.REQUESTER

        assert (
            ConnRecord.State.get(ConnRecord.State.RESPONSE) is ConnRecord.State.RESPONSE
        )

        assert ConnRecord.Role.REQUESTER.flip() is ConnRecord.Role.RESPONDER
        assert ConnRecord.Role.get(
            ConnRecord.Role.REQUESTER.rfc160
        ) is ConnRecord.Role.get(ConnRecord.Role.REQUESTER.rfc23)
        assert ConnRecord.Role.REQUESTER == ConnRecord.Role.REQUESTER.rfc160  # check ==
        assert ConnRecord.Role.REQUESTER == ConnRecord.Role.REQUESTER.rfc23
        assert ConnRecord.Role.REQUESTER != ConnRecord.Role.RESPONDER.rfc23

    async def test_state_rfc23strict(self):
        for state in (
            ConnRecord.State.INIT,
            ConnRecord.State.ABANDONED,
            ConnRecord.State.COMPLETED,
        ):
            assert state.rfc23strict(their_role=None) == state.value[1]

        for state in (ConnRecord.State.INVITATION, ConnRecord.State.RESPONSE):
            assert (
                state.rfc23strict(their_role=ConnRecord.Role.REQUESTER)
                == f"{state.value[1]}-sent"
            )
            assert (
                state.rfc23strict(their_role=ConnRecord.Role.RESPONDER)
                == f"{state.value[1]}-received"
            )

        assert (
            ConnRecord.State.REQUEST.rfc23strict(their_role=ConnRecord.Role.REQUESTER)
            == f"{ConnRecord.State.REQUEST.value[1]}-received"
        )
        assert (
            ConnRecord.State.REQUEST.rfc23strict(their_role=ConnRecord.Role.RESPONDER)
            == f"{ConnRecord.State.REQUEST.value[1]}-sent"
        )

    async def test_save_retrieve_compare(self):
        record = ConnRecord(my_did=self.test_did)
        async with self.profile.session() as session:
            connection_id = await record.save(session)
            fetched = await ConnRecord.retrieve_by_id(session, connection_id)
            assert fetched and fetched == record

            bad_record = ConnRecord(my_did=None)
            bad_record._id = record._id
            bad_record.created_at = record.created_at
            bad_record.updated_at = record.updated_at
            assert bad_record != record

            record = ConnRecord(
                state=ConnRecord.State.INIT,  # exercise init State by enum
                my_did=self.test_did,
                their_role=ConnRecord.Role.REQUESTER,  # exercise init Role by enum
            )
            connection_id = await record.save(session)
            fetched = await ConnRecord.retrieve_by_id(session, connection_id)
            assert fetched and fetched == record
            assert fetched.state is ConnRecord.State.INIT.rfc160
            assert ConnRecord.State.get(fetched.state) is ConnRecord.State.INIT
            assert fetched.their_role is ConnRecord.Role.REQUESTER.rfc160
            assert ConnRecord.Role.get(fetched.their_role) is ConnRecord.Role.REQUESTER

            record160 = ConnRecord(
                state=ConnRecord.State.INIT.rfc23,
                my_did=self.test_did,
                their_role=ConnRecord.Role.REQUESTER.rfc23,
            )
            record160._id = record._id
            record160.created_at = record.created_at
            record160.updated_at = record.updated_at
            assert record160 == record

    async def test_retrieve_by_did(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.COMPLETED.rfc23,
            )
            await record.save(session)
            result = await ConnRecord.retrieve_by_did(
                session=session,
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_role=ConnRecord.Role.RESPONDER.rfc160,
            )
            assert result == record

    async def test_retrieve_by_did_peer_4_by_long(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_did_peer_4_a,
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.COMPLETED.rfc23,
            )
            await record.save(session)
            result = await ConnRecord.retrieve_by_did_peer_4(
                session=session,
                my_did=self.test_did,
                their_did_long=self.test_did_peer_4_a,
                their_role=ConnRecord.Role.RESPONDER.rfc160,
            )
            assert result == record

    async def test_retrieve_by_did_peer_4_by_short(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_did_peer_4_short_b,
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.COMPLETED.rfc23,
            )
            await record.save(session)
            result = await ConnRecord.retrieve_by_did_peer_4(
                session=session,
                my_did=self.test_did,
                their_did_short=self.test_did_peer_4_short_b,
                their_role=ConnRecord.Role.RESPONDER.rfc160,
            )
            assert result == record

    async def test_retrieve_by_did_peer_4_by_either(self):
        async with self.profile.session() as session:
            record_short = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_did_peer_4_short_a,
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.COMPLETED.rfc23,
            )
            await record_short.save(session)
            record_long = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_did_peer_4_b,
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.COMPLETED.rfc23,
            )
            await record_long.save(session)

            result = await ConnRecord.retrieve_by_did_peer_4(
                session=session,
                my_did=self.test_did,
                their_did_short=self.test_did_peer_4_short_a,
                their_did_long=self.test_did_peer_4_a,
                their_role=ConnRecord.Role.RESPONDER.rfc160,
            )
            assert result == record_short
            result = await ConnRecord.retrieve_by_did_peer_4(
                session=session,
                my_did=self.test_did,
                their_did_short=self.test_did_peer_4_short_b,
                their_did_long=self.test_did_peer_4_b,
                their_role=ConnRecord.Role.RESPONDER.rfc160,
            )
            assert result == record_long

    async def test_from_storage_with_initiator_old(self):
        record = ConnRecord(my_did=self.test_did, state=ConnRecord.State.COMPLETED)
        ser = record.serialize()
        ser["initiator"] = "self"  # old-style ConnectionRecord
        ConnRecord.from_storage("conn-id", ser)

    async def test_retrieve_by_invitation_key(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_role=ConnRecord.Role.RESPONDER.rfc160,
                state=ConnRecord.State.INVITATION.rfc23,
                invitation_key="dummy",
            )
            await record.save(session)
            result = await ConnRecord.retrieve_by_invitation_key(
                session=session,
                invitation_key="dummy",
                their_role=ConnRecord.Role.RESPONDER.rfc23,
            )
            assert result == record
            with self.assertRaises(StorageNotFoundError):
                await ConnRecord.retrieve_by_invitation_key(
                    session=session,
                    invitation_key="dummy",
                    their_role=ConnRecord.Role.REQUESTER.rfc23,
                )

    async def test_retrieve_by_invitation_msg_id(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_role=ConnRecord.Role.RESPONDER.rfc160,
                state=ConnRecord.State.INVITATION.rfc160,
                invitation_msg_id="test123",
            )
            await record.save(session)
            result = await ConnRecord.retrieve_by_invitation_msg_id(
                session=session,
                invitation_msg_id="test123",
                their_role=ConnRecord.Role.RESPONDER.rfc160,
            )
            assert result
            assert result == record
            result = await ConnRecord.retrieve_by_invitation_msg_id(
                session=session,
                invitation_msg_id="test123",
                their_role=ConnRecord.Role.REQUESTER.rfc160,
            )
            assert not result

    async def test_find_existing_connection(self):
        async with self.profile.session() as session:
            record_a = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_role=ConnRecord.Role.RESPONDER.rfc160,
                state=ConnRecord.State.COMPLETED.rfc160,
                invitation_msg_id="test123",
                their_public_did="test_did_1",
            )
            await record_a.save(session)
            record_b = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_role=ConnRecord.Role.RESPONDER.rfc160,
                state=ConnRecord.State.INVITATION.rfc160,
                invitation_msg_id="test123",
                their_public_did="test_did_1",
            )
            await record_b.save(session)
            record_c = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_role=ConnRecord.Role.RESPONDER.rfc160,
                state=ConnRecord.State.COMPLETED.rfc160,
                invitation_msg_id="test123",
            )
            await record_c.save(session)
            result = await ConnRecord.find_existing_connection(
                session=session,
                their_public_did="test_did_1",
            )
            assert result
            assert result.state == "active"
            assert result.their_public_did == "test_did_1"

    async def test_retrieve_by_request_id(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.COMPLETED.rfc23,
                request_id="abc123",
            )
            await record.save(session)
            result = await ConnRecord.retrieve_by_request_id(
                session=session, request_id="abc123"
            )
            assert result == record

    async def test_completed_is_ready(self):
        async with self.profile.session() as session:
            record = ConnRecord(my_did=self.test_did, state=ConnRecord.State.COMPLETED)
            connection_id = await record.save(session)
            fetched = await ConnRecord.retrieve_by_id(session, connection_id)

        assert fetched.is_ready is True

    async def test_response_is_ready(self):
        async with self.profile.session() as session:
            record = ConnRecord(my_did=self.test_did, state=ConnRecord.State.RESPONSE)
            connection_id = await record.save(session)
            fetched = await ConnRecord.retrieve_by_id(session, connection_id)

            assert fetched.is_ready is True

    async def test_request_is_not_ready(self):
        async with self.profile.session() as session:
            record = ConnRecord(my_did=self.test_did, state=ConnRecord.State.REQUEST)
            connection_id = await record.save(session)
            fetched = await ConnRecord.retrieve_by_id(session, connection_id)

            assert fetched.is_ready is False

    async def test_invitation_is_not_multi_use(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                state=ConnRecord.State.INVITATION.rfc23,
                invitation_mode=ConnRecord.INVITATION_MODE_ONCE,
            )
            connection_id = await record.save(session)
            fetched = await ConnRecord.retrieve_by_id(session, connection_id)

        assert fetched.is_multiuse_invitation is False

    async def test_invitation_is_multi_use(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                state=ConnRecord.State.INVITATION.rfc23,
                invitation_mode=ConnRecord.INVITATION_MODE_MULTI,
            )
            connection_id = await record.save(session)
            fetched = await ConnRecord.retrieve_by_id(session, connection_id)

            assert fetched.is_multiuse_invitation is True

    async def test_attach_retrieve_invitation(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                state=ConnRecord.State.INVITATION.rfc23,
            )
            await record.save(session)

            service = Service(
                _id="asdf",
                _type="did-communication",
                recipient_keys=[self.test_did_key.did],
                service_endpoint="http://localhost:8999",
            )
            invi = InvitationMessage(
                handshake_protocols=["didexchange/1.1"],
                services=[service],
                label="abc123",
            )
            await record.attach_invitation(session, invi)
            retrieved = await record.retrieve_invitation(session)
            assert isinstance(retrieved, InvitationMessage)

    async def test_attach_retrieve_request(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                state=ConnRecord.State.INVITATION.rfc23,
            )
            await record.save(session)

            req = DIDXRequest(
                did=self.test_did,
                label="abc123",
            )
            await record.attach_request(session, req)
            retrieved = await record.retrieve_request(session)
            assert isinstance(retrieved, DIDXRequest)

    async def test_attach_request_abstain_on_alien_deco(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
                state=ConnRecord.State.INVITATION.rfc23,
            )
            await record.save(session)

            req = DIDXRequest(
                did=self.test_did,
                label="abc123",
            )
            ser = req.serialize()
            ser["~alien"] = [
                {"nickname": "profile-image", "data": {"links": ["face.png"]}}
            ]
            alien_req = DIDXRequest.deserialize(ser)
            await record.attach_request(session, alien_req)
            alien_ser = alien_req.serialize()
            assert "~alien" in alien_ser

            ser["~alien"] = None
            alien_req = DIDXRequest.deserialize(ser)
            await record.attach_request(session, alien_req)
            alien_ser = alien_req.serialize()
            assert "~alien" not in alien_ser

    async def test_ser_rfc23_state_present(self):
        record = ConnRecord(
            state=ConnRecord.State.INVITATION,
            my_did=self.test_did,
            their_role=ConnRecord.Role.REQUESTER,
        )
        ser = record.serialize()
        assert ser["rfc23_state"] == f"{ConnRecord.State.INVITATION.value[1]}-sent"

    async def test_deser_old_style_record(self):
        record = ConnRecord(
            state=ConnRecord.State.INIT,
            my_did=self.test_did,
            their_role=ConnRecord.Role.REQUESTER,
        )
        ser = record.serialize()
        ser["initiator"] = "self"  # redundant vs. role as per RFC 160 or RFC 23
        deser = ConnRecord.deserialize(ser)
        reser = deser.serialize()
        assert "initiator" not in reser

    async def test_deserialize_connection_protocol(self):
        record = ConnRecord(
            state=ConnRecord.State.INIT,
            my_did=self.test_did,
            their_role=ConnRecord.Role.REQUESTER,
            connection_protocol="didexchange/1.0",
        )
        ser = record.serialize()
        deser = ConnRecord.deserialize(ser)
        assert deser.connection_protocol == "didexchange/1.0"

    async def test_metadata_set_get(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
            )
            await record.save(session)
            await record.metadata_set(session, "key", {"test": "value"})
            retrieved = await record.metadata_get(session, "key")
            assert retrieved == {"test": "value"}

    async def test_metadata_set_get_str(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
            )
            await record.save(session)
            await record.metadata_set(session, "key", "value")
            retrieved = await record.metadata_get(session, "key")
            assert retrieved == "value"

    async def test_metadata_set_update_get(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
            )
            await record.save(session)
            await record.metadata_set(session, "key", {"test": "value"})
            await record.metadata_set(session, "key", {"test": "updated"})
            retrieved = await record.metadata_get(session, "key")
            assert retrieved == {"test": "updated"}

    async def test_metadata_get_without_set_is_none(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
            )
            await record.save(session)
            assert await record.metadata_get(session, "key") is None

    async def test_metadata_get_default(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
            )
            await record.save(session)
            assert await record.metadata_get(session, "key", {"test": "default"}) == {
                "test": "default"
            }

    async def test_metadata_set_delete_get_is_none(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
            )
            await record.save(session)
            await record.metadata_set(session, "key", {"test": "value"})
            await record.metadata_delete(session, "key")
            assert await record.metadata_get(session, "key") is None

    async def test_metadata_delete_without_set_raise_error(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
            )
            await record.save(session)
            with self.assertRaises(KeyError):
                await record.metadata_delete(session, "key")

    async def test_metadata_get_all(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
            )
            await record.save(session)
            await record.metadata_set(session, "key", {"test": "value"})
            await record.metadata_set(session, "key", {"test": "updated"})
            await record.metadata_set(session, "other", {"test": "other"})
            retrieved = await record.metadata_get_all(session)
            assert retrieved == {"key": {"test": "updated"}, "other": {"test": "other"}}

    async def test_metadata_get_all_without_set_is_empty(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
            )
            await record.save(session)
            assert await record.metadata_get_all(session) == {}

    async def test_delete_conn_record_deletes_metadata(self):
        async with self.profile.session() as session:
            record = ConnRecord(
                my_did=self.test_did,
            )
            await record.save(session)
            await record.metadata_set(session, "key", {"test": "value"})
            await record.delete_record(session)
            storage = session.inject(BaseStorage)
            assert (
                await storage.find_all_records(
                    ConnRecord.RECORD_TYPE_METADATA,
                    {"connection_id": record.connection_id},
                )
                == []
            )



import json
from copy import deepcopy
from time import time
from unittest import IsolatedAsyncioTestCase

import pytest
from anoncreds import CredentialDefinition
from marshmallow import ValidationError

from .......anoncreds.holder import AnonCredsHolder
from .......anoncreds.issuer import AnonCredsIssuer
from .......anoncreds.models.credential_definition import (
    CredDef,
    CredDefValue,
    CredDefValuePrimary,
)
from .......anoncreds.registry import AnonCredsRegistry
from .......anoncreds.revocation import AnonCredsRevocationRegistryFullError
from .......cache.base import BaseCache
from .......cache.in_memory import InMemoryCache
from .......config.provider import ClassProvider
from .......indy.credx.issuer import CATEGORY_CRED_DEF
from .......ledger.base import BaseLedger
from .......ledger.multiple_ledger.ledger_requests_executor import (
    IndyLedgerRequestsExecutor,
)
from .......messaging.credential_definitions.util import CRED_DEF_SENT_RECORD_TYPE
from .......messaging.decorators.attach_decorator import AttachDecorator
from .......multitenant.base import BaseMultitenantManager
from .......multitenant.manager import MultitenantManager
from .......storage.record import StorageRecord
from .......tests import mock
from .......utils.testing import create_test_profile
from ....message_types import (
    ATTACHMENT_FORMAT,
    CRED_20_ISSUE,
    CRED_20_OFFER,
    CRED_20_PROPOSAL,
    CRED_20_REQUEST,
)
from ....messages.cred_format import V20CredFormat
from ....messages.cred_issue import V20CredIssue
from ....messages.cred_offer import V20CredOffer
from ....messages.cred_proposal import V20CredProposal
from ....messages.cred_request import V20CredRequest
from ....messages.inner.cred_preview import V20CredAttrSpec, V20CredPreview
from ....models.cred_ex_record import V20CredExRecord
from ....models.detail.anoncreds import V20CredExRecordAnonCreds
from ...handler import V20CredFormatError
from .. import handler as test_module
from ..handler import LOGGER as ANONCREDS_LOGGER
from ..handler import AnonCredsCredFormatHandler

TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "bc-reg"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
SCHEMA = {
    "ver": "1.0",
    "id": SCHEMA_ID,
    "name": SCHEMA_NAME,
    "version": "1.0",
    "attrNames": ["legalName", "jurisdictionId", "incorporationDate"],
    "seqNo": SCHEMA_TXN,
}
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:tag1"
CRED_DEF = {
    "ver": "1.0",
    "id": CRED_DEF_ID,
    "schemaId": SCHEMA_TXN,
    "type": "CL",
    "tag": "tag1",
    "value": {
        "primary": {
            "n": "...",
            "s": "...",
            "r": {
                "master_secret": "...",
                "legalName": "...",
                "jurisdictionId": "...",
                "incorporationDate": "...",
            },
            "rctxt": "...",
            "z": "...",
        },
        "revocation": {
            "g": "1 ...",
            "g_dash": "1 ...",
            "h": "1 ...",
            "h0": "1 ...",
            "h1": "1 ...",
            "h2": "1 ...",
            "htilde": "1 ...",
            "h_cap": "1 ...",
            "u": "1 ...",
            "pk": "1 ...",
            "y": "1 ...",
        },
    },
}
REV_REG_DEF_TYPE = "CL_ACCUM"
REV_REG_ID = f"{TEST_DID}:4:{CRED_DEF_ID}:{REV_REG_DEF_TYPE}:tag1"
TAILS_DIR = "/tmp/indy/revocation/tails_files"
TAILS_HASH = "8UW1Sz5cqoUnK9hqQk7nvtKK65t7Chu3ui866J23sFyJ"
TAILS_LOCAL = f"{TAILS_DIR}/{TAILS_HASH}"
REV_REG_DEF = {
    "ver": "1.0",
    "id": REV_REG_ID,
    "revocDefType": "CL_ACCUM",
    "tag": "tag1",
    "credDefId": CRED_DEF_ID,
    "value": {
        "issuanceType": "ISSUANCE_ON_DEMAND",
        "maxCredNum": 5,
        "publicKeys": {"accumKey": {"z": "1 ..."}},
        "tailsHash": TAILS_HASH,
        "tailsLocation": TAILS_LOCAL,
    },
}
ANONCREDS_OFFER = {
    "schema_id": SCHEMA_ID,
    "cred_def_id": CRED_DEF_ID,
    "key_correctness_proof": {
        "c": "123467890",
        "xz_cap": "12345678901234567890",
        "xr_cap": [
            [
                "remainder",
                "1234567890",
            ],
            [
                "number",
                "12345678901234",
            ],
            [
                "master_secret",
                "12345678901234",
            ],
        ],
    },
    "nonce": "1234567890",
}
ANONCREDS_CRED_REQ = {
    "entropy": TEST_DID,
    "cred_def_id": CRED_DEF_ID,
    "blinded_ms": {
        "u": "12345",
        "ur": "1 123467890ABCDEF",
        "hidden_attributes": ["master_secret"],
        "committed_attributes": {},
    },
    "blinded_ms_correctness_proof": {
        "c": "77777",
        "v_dash_cap": "12345678901234567890",
        "m_caps": {"master_secret": "271283714"},
        "r_caps": {},
    },
    "nonce": "9876543210",
}
ANONCREDS_CRED = {
    "schema_id": SCHEMA_ID,
    "cred_def_id": CRED_DEF_ID,
    "rev_reg_id": REV_REG_ID,
    "values": {
        "legalName": {
            "raw": "The Original House of Pies",
            "encoded": "108156129846915621348916581250742315326283968964",
        },
        "busId": {"raw": "11155555", "encoded": "11155555"},
        "jurisdictionId": {"raw": "1", "encoded": "1"},
        "incorporationDate": {
            "raw": "2021-01-01",
            "encoded": "121381685682968329568231",
        },
        "pic": {"raw": "cG90YXRv", "encoded": "125362825623562385689562"},
    },
    "signature": {
        "p_credential": {
            "m_2": "13683295623862356",
            "a": "1925723185621385238953",
            "e": "253516862326",
            "v": "26890295622385628356813632",
        },
        "r_credential": {
            "sigma": "1 00F81D",
            "c": "158698926BD09866E",
            "vr_prime_prime": "105682396DDF1A",
            "witness_signature": {"sigma_i": "1 ...", "u_i": "1 ...", "g_i": "1 ..."},
            "g_i": "1 ...",
            "i": 1,
            "m2": "862186285926592362384FA97FF3A4AB",
        },
    },
    "signature_correctness_proof": {
        "se": "10582965928638296868123",
        "c": "2816389562839651",
    },
    "rev_reg": {"accum": "21 ..."},
    "witness": {"omega": "21 ..."},
}


class TestV20AnonCredsCredFormatHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            {
                "wallet.type": "askar-anoncreds",
            }
        )
        self.context = self.profile.context

        # Context
        self.cache = InMemoryCache()
        self.profile.context.injector.bind_instance(BaseCache, self.cache)

        # Issuer
        self.issuer = mock.MagicMock(AnonCredsIssuer, autospec=True)
        self.profile.context.injector.bind_instance(AnonCredsIssuer, self.issuer)

        # Holder
        self.holder = mock.MagicMock(AnonCredsHolder, autospec=True)
        self.profile.context.injector.bind_instance(AnonCredsHolder, self.holder)

        # AnonCreds registry
        self.profile.context.injector.bind_instance(
            AnonCredsRegistry, AnonCredsRegistry()
        )
        registry = self.profile.context.inject_or(AnonCredsRegistry)
        legacy_indy_registry = ClassProvider(
            "acapy_agent.anoncreds.default.legacy_indy.registry.LegacyIndyRegistry",
            # supported_identifiers=[],
            # method_name="",
        ).provide(self.profile.context.settings, self.profile.context.injector)
        await legacy_indy_registry.setup(self.profile.context)
        registry.register(legacy_indy_registry)

        # Ledger
        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_schema = mock.CoroutineMock(return_value=SCHEMA)
        self.ledger.get_credential_definition = mock.CoroutineMock(return_value=CRED_DEF)
        self.ledger.get_revoc_reg_def = mock.CoroutineMock(return_value=REV_REG_DEF)
        self.ledger.__aenter__ = mock.CoroutineMock(return_value=self.ledger)
        self.ledger.credential_definition_id2schema_id = mock.CoroutineMock(
            return_value=SCHEMA_ID
        )
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor,
            mock.MagicMock(
                get_ledger_for_identifier=mock.CoroutineMock(
                    return_value=(None, self.ledger)
                )
            ),
        )
        self.handler = AnonCredsCredFormatHandler(self.profile)
        assert self.handler.profile

    async def test_validate_fields(self):
        # Test correct data
        self.handler.validate_fields(CRED_20_PROPOSAL, {"cred_def_id": CRED_DEF_ID})
        self.handler.validate_fields(CRED_20_OFFER, ANONCREDS_OFFER)
        self.handler.validate_fields(CRED_20_REQUEST, ANONCREDS_CRED_REQ)
        self.handler.validate_fields(CRED_20_ISSUE, ANONCREDS_CRED)

        # test incorrect proposal
        with self.assertRaises(ValidationError):
            self.handler.validate_fields(
                CRED_20_PROPOSAL, {"some_random_key": "some_random_value"}
            )

        # test incorrect offer
        with self.assertRaises(ValidationError):
            offer = ANONCREDS_OFFER.copy()
            offer.pop("nonce")
            self.handler.validate_fields(CRED_20_OFFER, offer)

        # test incorrect request
        with self.assertRaises(ValidationError):
            req = ANONCREDS_CRED_REQ.copy()
            req.pop("nonce")
            self.handler.validate_fields(CRED_20_REQUEST, req)

        # test incorrect cred
        with self.assertRaises(ValidationError):
            cred = ANONCREDS_CRED.copy()
            cred.pop("schema_id")
            self.handler.validate_fields(CRED_20_ISSUE, cred)

    async def test_get_indy_detail_record(self):
        cred_ex_id = "dummy"
        details_indy = [
            V20CredExRecordAnonCreds(
                cred_ex_id=cred_ex_id,
                rev_reg_id="rr-id",
                cred_rev_id="0",
            ),
            V20CredExRecordAnonCreds(
                cred_ex_id=cred_ex_id,
                rev_reg_id="rr-id",
                cred_rev_id="1",
            ),
        ]
        async with self.profile.session() as session:
            await details_indy[0].save(session)
            await details_indy[1].save(session)  # exercise logger warning on get()

        with mock.patch.object(
            ANONCREDS_LOGGER, "warning", mock.MagicMock()
        ) as mock_warning:
            assert await self.handler.get_detail_record(cred_ex_id) in details_indy
            mock_warning.assert_called_once()

    async def test_check_uniqueness(self):
        with mock.patch.object(
            self.handler.format.detail,
            "query_by_cred_ex_id",
            mock.CoroutineMock(),
        ) as mock_indy_query:
            mock_indy_query.return_value = []
            await self.handler._check_uniqueness("dummy-cx-id")

        with mock.patch.object(
            self.handler.format.detail,
            "query_by_cred_ex_id",
            mock.CoroutineMock(),
        ) as mock_indy_query:
            mock_indy_query.return_value = [mock.MagicMock()]
            with self.assertRaises(V20CredFormatError) as context:
                await self.handler._check_uniqueness("dummy-cx-id")
            assert "detail record already exists" in str(context.exception)

    async def test_create_proposal(self):
        cred_ex_record = mock.MagicMock()
        proposal_data = {"schema_id": SCHEMA_ID}

        (cred_format, attachment) = await self.handler.create_proposal(
            cred_ex_record, proposal_data
        )

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == proposal_data

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_create_proposal_none(self):
        cred_ex_record = mock.MagicMock()
        proposal_data = None

        (cred_format, attachment) = await self.handler.create_proposal(
            cred_ex_record, proposal_data
        )

        # assert content of attachment is proposal data
        assert attachment.content == {}

    async def test_receive_proposal(self):
        cred_ex_record = mock.MagicMock()
        cred_proposal_message = mock.MagicMock()

        # Not much to assert. Receive proposal doesn't do anything
        await self.handler.receive_proposal(cred_ex_record, cred_proposal_message)

    async def test_create_offer_cant_find_schema_in_wallet_or_data_registry(self):
        with self.assertRaises(V20CredFormatError):
            await self.handler.create_offer(
                V20CredProposal(
                    formats=[
                        V20CredFormat(
                            attach_id="0",
                            format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                                V20CredFormat.Format.ANONCREDS.api
                            ],
                        )
                    ],
                    filters_attach=[
                        AttachDecorator.data_base64(
                            {"cred_def_id": CRED_DEF_ID}, ident="0"
                        )
                    ],
                )
            )

    @mock.patch.object(
        AnonCredsRegistry,
        "get_schema",
        mock.CoroutineMock(
            return_value=mock.MagicMock(schema=mock.MagicMock(attr_names=["score"]))
        ),
    )
    @mock.patch.object(
        AnonCredsIssuer,
        "create_credential_offer",
        mock.CoroutineMock(return_value=json.dumps(ANONCREDS_OFFER)),
    )
    @mock.patch.object(
        CredentialDefinition,
        "load",
        mock.MagicMock(to_dict=mock.MagicMock(return_value={"schemaId": SCHEMA_ID})),
    )
    async def test_create_offer(self):
        self.issuer.create_credential_offer = mock.CoroutineMock({})
        # With a schema_id
        await self.handler.create_offer(
            V20CredProposal(
                credential_preview=V20CredPreview(
                    attributes=(V20CredAttrSpec(name="score", value="0"),)
                ),
                formats=[
                    V20CredFormat(
                        attach_id="0",
                        format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                            V20CredFormat.Format.ANONCREDS.api
                        ],
                    )
                ],
                filters_attach=[
                    AttachDecorator.data_base64(
                        {"cred_def_id": CRED_DEF_ID, "schema_id": SCHEMA_ID}, ident="0"
                    )
                ],
            )
        )
        # Only with cred_def_id
        async with self.profile.session() as session:
            await session.handle.insert(
                CATEGORY_CRED_DEF,
                CRED_DEF_ID,
                CredDef(
                    issuer_id=TEST_DID,
                    schema_id=SCHEMA_ID,
                    tag="tag",
                    type="CL",
                    value=CredDefValue(
                        primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")
                    ),
                ).to_json(),
                tags={},
            )
        await self.handler.create_offer(
            V20CredProposal(
                credential_preview=V20CredPreview(
                    attributes=(V20CredAttrSpec(name="score", value="0"),)
                ),
                formats=[
                    V20CredFormat(
                        attach_id="0",
                        format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                            V20CredFormat.Format.ANONCREDS.api
                        ],
                    )
                ],
                filters_attach=[
                    AttachDecorator.data_base64({"cred_def_id": CRED_DEF_ID}, ident="0")
                ],
            )
        )
        # Wrong attribute name
        with self.assertRaises(V20CredFormatError):
            await self.handler.create_offer(
                V20CredProposal(
                    credential_preview=V20CredPreview(
                        attributes=(V20CredAttrSpec(name="wrong", value="0"),)
                    ),
                    formats=[
                        V20CredFormat(
                            attach_id="0",
                            format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                                V20CredFormat.Format.ANONCREDS.api
                            ],
                        )
                    ],
                    filters_attach=[
                        AttachDecorator.data_base64(
                            {"cred_def_id": CRED_DEF_ID, "schema_id": SCHEMA_ID},
                            ident="0",
                        )
                    ],
                )
            )

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_offer_no_cache(self):
        schema_id_parts = SCHEMA_ID.split(":")

        cred_preview = V20CredPreview(
            attributes=(
                V20CredAttrSpec(name="legalName", value="value"),
                V20CredAttrSpec(name="jurisdictionId", value="value"),
                V20CredAttrSpec(name="incorporationDate", value="value"),
            )
        )

        cred_proposal = V20CredProposal(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            filters_attach=[
                AttachDecorator.data_base64({"cred_def_id": CRED_DEF_ID}, ident="0")
            ],
        )

        cred_def_record = StorageRecord(
            CRED_DEF_SENT_RECORD_TYPE,
            CRED_DEF_ID,
            {
                "schema_id": SCHEMA_ID,
                "schema_issuer_did": schema_id_parts[0],
                "schema_name": schema_id_parts[-2],
                "schema_version": schema_id_parts[-1],
                "issuer_did": TEST_DID,
                "cred_def_id": CRED_DEF_ID,
                "epoch": str(int(time())),
            },
        )

        # Remove cache from injection context
        self.context.injector.clear_binding(BaseCache)

        await self.session.storage.add_record(cred_def_record)

        self.issuer.create_credential_offer = mock.CoroutineMock(
            return_value=json.dumps(ANONCREDS_OFFER)
        )

        (cred_format, attachment) = await self.handler.create_offer(cred_proposal)

        self.issuer.create_credential_offer.assert_called_once_with(CRED_DEF_ID)

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == ANONCREDS_OFFER

        # assert data is encoded as base64
        assert attachment.data.base64

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_offer_attr_mismatch(self):
        schema_id_parts = SCHEMA_ID.split(":")

        cred_preview = V20CredPreview(
            attributes=(  # names have spaces instead of camel case
                V20CredAttrSpec(name="legal name", value="value"),
                V20CredAttrSpec(name="jurisdiction id", value="value"),
                V20CredAttrSpec(name="incorporation date", value="value"),
            )
        )

        cred_proposal = V20CredProposal(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            filters_attach=[
                AttachDecorator.data_base64({"cred_def_id": CRED_DEF_ID}, ident="0")
            ],
        )
        self.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )

        cred_def_record = StorageRecord(
            CRED_DEF_SENT_RECORD_TYPE,
            CRED_DEF_ID,
            {
                "schema_id": SCHEMA_ID,
                "schema_issuer_did": schema_id_parts[0],
                "schema_name": schema_id_parts[-2],
                "schema_version": schema_id_parts[-1],
                "issuer_did": TEST_DID,
                "cred_def_id": CRED_DEF_ID,
                "epoch": str(int(time())),
            },
        )
        await self.session.storage.add_record(cred_def_record)

        self.issuer.create_credential_offer = mock.CoroutineMock(
            return_value=json.dumps(ANONCREDS_OFFER)
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor,
            "get_ledger_for_identifier",
            mock.CoroutineMock(return_value=(None, self.ledger)),
        ):
            with self.assertRaises(V20CredFormatError):
                await self.handler.create_offer(cred_proposal)

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_offer_no_matching_sent_cred_def(self):
        cred_proposal = V20CredProposal(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            filters_attach=[AttachDecorator.data_base64({}, ident="0")],
        )

        self.issuer.create_credential_offer = mock.CoroutineMock(
            return_value=json.dumps(ANONCREDS_OFFER)
        )

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.create_offer(cred_proposal)
        assert "Issuer has no operable cred def" in str(context.exception)

    async def test_receive_offer(self):
        cred_ex_record = mock.MagicMock()
        cred_offer_message = mock.MagicMock()

        # Not much to assert. Receive offer doesn't do anything
        await self.handler.receive_offer(cred_ex_record, cred_offer_message)

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_create_request(self):
        holder_did = "did"

        cred_offer = V20CredOffer(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(ANONCREDS_OFFER, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-id",
            state=V20CredExRecord.STATE_OFFER_RECEIVED,
            cred_offer=cred_offer.serialize(),
        )

        cred_def = {"cred": "def"}
        self.ledger.get_credential_definition = mock.CoroutineMock(return_value=cred_def)

        cred_req_meta = {}
        self.holder.create_credential_request = mock.CoroutineMock(
            return_value=(json.dumps(ANONCREDS_CRED_REQ), json.dumps(cred_req_meta))
        )

        (cred_format, attachment) = await self.handler.create_request(
            cred_ex_record, {"holder_did": holder_did}
        )

        self.holder.create_credential_request.assert_called_once_with(
            ANONCREDS_OFFER, cred_def, holder_did
        )

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == ANONCREDS_CRED_REQ

        # assert data is encoded as base64
        assert attachment.data.base64

        # cover case with cache (change ID to prevent already exists error)
        cred_ex_record._id = "dummy-id2"
        await self.handler.create_request(cred_ex_record, {"holder_did": holder_did})

        # cover case with no cache in injection context
        self.context.injector.clear_binding(BaseCache)
        cred_ex_record._id = "dummy-id3"
        self.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor,
            "get_ledger_for_identifier",
            mock.CoroutineMock(return_value=(None, self.ledger)),
        ):
            await self.handler.create_request(cred_ex_record, {"holder_did": holder_did})

    async def test_create_request_bad_state(self):
        cred_ex_record = V20CredExRecord(state=V20CredExRecord.STATE_OFFER_SENT)

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.create_request(cred_ex_record)
        assert (
            "AnonCreds issue credential format cannot start from credential request"
            in str(context.exception)
        )

        cred_ex_record.state = None

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.create_request(cred_ex_record)
        assert (
            "AnonCreds issue credential format cannot start from credential request"
            in str(context.exception)
        )

    async def test_create_request_not_unique_x(self):
        cred_ex_record = V20CredExRecord(state=V20CredExRecord.STATE_OFFER_RECEIVED)

        with mock.patch.object(
            self.handler, "_check_uniqueness", mock.CoroutineMock()
        ) as mock_unique:
            mock_unique.side_effect = (
                V20CredFormatError("indy detail record already exists"),
            )

            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.create_request(cred_ex_record)

            assert "indy detail record already exists" in str(context.exception)

    async def test_receive_request(self):
        cred_ex_record = mock.MagicMock()
        cred_request_message = mock.MagicMock()

        # Not much to assert. Receive request doesn't do anything
        await self.handler.receive_request(cred_ex_record, cred_request_message)

    async def test_receive_request_no_offer(self):
        cred_ex_record = mock.MagicMock(cred_offer=None)
        cred_request_message = mock.MagicMock()

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.receive_request(cred_ex_record, cred_request_message)

        assert (
            "AnonCreds issue credential format cannot start from credential request"
            in str(context.exception)
        )

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_issue_credential_revocable(self):
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(ANONCREDS_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(ANONCREDS_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        cred_rev_id = "1000"
        self.issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(ANONCREDS_CRED), cred_rev_id)
        )

        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                return_value=(
                    mock.MagicMock(  # active_rev_reg_rec
                        revoc_reg_id=REV_REG_ID,
                    ),
                    mock.MagicMock(  # rev_reg
                        tails_local_path="dummy-path",
                        get_or_fetch_local_tails_path=(mock.CoroutineMock()),
                        max_creds=10,
                    ),
                )
            )

            (cred_format, attachment) = await self.handler.issue_credential(
                cred_ex_record, retries=1
            )

            self.issuer.create_credential.assert_called_once_with(
                SCHEMA,
                ANONCREDS_OFFER,
                ANONCREDS_CRED_REQ,
                attr_values,
                REV_REG_ID,
                "dummy-path",
            )

            # assert identifier match
            assert cred_format.attach_id == self.handler.format.api == attachment.ident

            # assert content of attachment is proposal data
            assert attachment.content == ANONCREDS_CRED

            # assert data is encoded as base64
            assert attachment.data.base64

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_issue_credential_non_revocable(self):
        CRED_DEF_NR = deepcopy(CRED_DEF)
        CRED_DEF_NR["value"]["revocation"] = None
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(ANONCREDS_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(ANONCREDS_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        self.issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(ANONCREDS_CRED), None)
        )
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value=CRED_DEF_NR
        )
        self.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor,
            "get_ledger_for_identifier",
            mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
        ):
            (cred_format, attachment) = await self.handler.issue_credential(
                cred_ex_record, retries=0
            )

            self.issuer.create_credential.assert_called_once_with(
                SCHEMA,
                ANONCREDS_OFFER,
                ANONCREDS_CRED_REQ,
                attr_values,
                None,
                None,
            )

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == ANONCREDS_CRED

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_issue_credential_not_unique_x(self):
        cred_ex_record = V20CredExRecord(state=V20CredExRecord.STATE_REQUEST_RECEIVED)

        with mock.patch.object(
            self.handler, "_check_uniqueness", mock.CoroutineMock()
        ) as mock_unique:
            mock_unique.side_effect = (
                V20CredFormatError("indy detail record already exists"),
            )

            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.issue_credential(cred_ex_record)

            assert "indy detail record already exists" in str(context.exception)

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_issue_credential_no_active_rr_no_retries(self):
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_rev_id = "1"

        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(ANONCREDS_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(ANONCREDS_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        self.issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(ANONCREDS_CRED), cred_rev_id)
        )

        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                return_value=()
            )
            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.issue_credential(cred_ex_record, retries=0)
            assert "has no active revocation registry" in str(context.exception)

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_issue_credential_no_active_rr_retry(self):
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_rev_id = "1"

        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(ANONCREDS_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(ANONCREDS_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        self.issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(ANONCREDS_CRED), cred_rev_id)
        )

        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                side_effect=[
                    None,
                    (
                        mock.MagicMock(  # active_rev_reg_rec
                            revoc_reg_id=REV_REG_ID,
                            set_state=mock.CoroutineMock(),
                        ),
                        mock.MagicMock(  # rev_reg
                            tails_local_path="dummy-path",
                            get_or_fetch_local_tails_path=(mock.CoroutineMock()),
                        ),
                    ),
                ]
            )

            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.issue_credential(cred_ex_record, retries=1)
            assert "has no active revocation registry" in str(context.exception)

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_issue_credential_rr_full(self):
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(ANONCREDS_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(ANONCREDS_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        self.issuer.create_credential = mock.AsyncMock(
            side_effect=AnonCredsRevocationRegistryFullError("Nope")
        )
        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                return_value=(
                    mock.MagicMock(  # active_rev_reg_rec
                        revoc_reg_id=REV_REG_ID,
                        set_state=mock.CoroutineMock(),
                    ),
                    mock.MagicMock(  # rev_reg
                        tails_local_path="dummy-path",
                        get_or_fetch_local_tails_path=(mock.CoroutineMock()),
                    ),
                )
            )

            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.issue_credential(cred_ex_record, retries=1)
            assert "has no active revocation registry" in str(context.exception)

    async def test_receive_credential(self):
        cred_ex_record = mock.MagicMock()
        cred_issue_message = mock.MagicMock()

        # Not much to assert. Receive credential doesn't do anything
        await self.handler.receive_credential(cred_ex_record, cred_issue_message)

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_store_credential(self):
        connection_id = "test_conn_id"
        attr_values = {
            "legalName": ["value", None],
            "jurisdictionId": ["value", None],
            "incorporationDate": ["value", None],
            "pic": ["cG90YXRv", "image/jpeg"],
        }
        cred_req_meta = {"req": "meta"}
        thread_id = "thread-id"

        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v[0], mime_type=v[1])
                for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(ANONCREDS_OFFER, ident="0")],
        )
        cred_offer.assign_thread_id(thread_id)
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(ANONCREDS_CRED_REQ, ident="0")],
        )
        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(ANONCREDS_CRED, ident="0")],
        )

        stored_cx_rec = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            connection_id=connection_id,
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            cred_issue=cred_issue.serialize(),
            initiator=V20CredExRecord.INITIATOR_EXTERNAL,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_CREDENTIAL_RECEIVED,
            thread_id=thread_id,
            auto_remove=True,
        )

        cred_id = "cred-id"

        self.holder.store_credential = mock.CoroutineMock(return_value=cred_id)
        stored_cred = {"stored": "cred"}
        self.holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(stored_cred)
        )

        with mock.patch.object(
            test_module, "RevocationRegistry", autospec=True
        ) as mock_rev_reg:
            mock_rev_reg.from_definition = mock.MagicMock(
                return_value=mock.MagicMock(
                    get_or_fetch_local_tails_path=mock.CoroutineMock()
                )
            )
            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.store_credential(stored_cx_rec, cred_id=cred_id)
            assert "No credential exchange " in str(context.exception)
        self.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch.object(
                test_module, "RevocationRegistry", autospec=True
            ) as mock_rev_reg,
            mock.patch.object(
                test_module.AnonCredsCredFormatHandler, "get_detail_record", autospec=True
            ) as mock_get_detail_record,
        ):
            mock_rev_reg.from_definition = mock.MagicMock(
                return_value=mock.MagicMock(
                    get_or_fetch_local_tails_path=mock.CoroutineMock()
                )
            )
            mock_get_detail_record.return_value = mock.MagicMock(
                cred_request_metadata=cred_req_meta,
                save=mock.CoroutineMock(),
            )

            self.ledger.get_credential_definition.reset_mock()
            await self.handler.store_credential(stored_cx_rec, cred_id=cred_id)

            self.ledger.get_credential_definition.assert_called_once_with(CRED_DEF_ID)

            self.holder.store_credential.assert_called_once_with(
                CRED_DEF,
                ANONCREDS_CRED,
                cred_req_meta,
                {"pic": "image/jpeg"},
                credential_id=cred_id,
                rev_reg_def=REV_REG_DEF,
            )

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_store_credential_holder_store_indy_error(self):
        connection_id = "test_conn_id"
        attr_values = {
            "legalName": ["value", None],
            "jurisdictionId": ["value", None],
            "incorporationDate": ["value", None],
            "pic": ["cG90YXRv", "image/jpeg"],
        }
        cred_req_meta = {"req": "meta"}
        thread_id = "thread-id"

        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v[0], mime_type=v[1])
                for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(ANONCREDS_OFFER, ident="0")],
        )
        cred_offer.assign_thread_id(thread_id)
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(ANONCREDS_CRED_REQ, ident="0")],
        )
        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.ANONCREDS.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(ANONCREDS_CRED, ident="0")],
        )

        stored_cx_rec = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            connection_id=connection_id,
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            cred_issue=cred_issue.serialize(),
            initiator=V20CredExRecord.INITIATOR_EXTERNAL,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_CREDENTIAL_RECEIVED,
            thread_id=thread_id,
            auto_remove=True,
        )

        cred_id = "cred-id"
        self.holder.store_credential = mock.AsyncMock(
            side_effect=test_module.AnonCredsHolderError("Problem", {"message": "Nope"})
        )

        with (
            mock.patch.object(
                test_module.AnonCredsCredFormatHandler, "get_detail_record", autospec=True
            ) as mock_get_detail_record,
            mock.patch.object(
                test_module.RevocationRegistry, "from_definition", mock.MagicMock()
            ) as mock_rev_reg,
        ):
            mock_get_detail_record.return_value = mock.MagicMock(
                cred_request_metadata=cred_req_meta,
                save=mock.CoroutineMock(),
            )
            mock_rev_reg.return_value = mock.MagicMock(
                get_or_fetch_local_tails_path=mock.CoroutineMock()
            )
            with self.assertRaises(test_module.AnonCredsHolderError) as context:
                await self.handler.store_credential(stored_cx_rec, cred_id)
            assert "Nope" in str(context.exception)



"""Ping handler."""

from .....messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from ..messages.ping import Ping
from ..messages.ping_response import PingResponse


class PingHandler(BaseHandler):
    """Ping handler class."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle ping message.

        Args:
            context: Request context
            responder: Responder used to reply

        """
        self._logger.debug(f"PingHandler called with context {context}")
        assert isinstance(context.message, Ping)

        self._logger.debug(
            "Received trust ping from: %s", context.message_receipt.sender_did
        )

        if not context.connection_ready:
            self._logger.info(
                "Connection not active, skipping ping response: %s",
                context.message_receipt.sender_did,
            )
            return

        if context.message.response_requested:
            reply = PingResponse()
            reply.assign_thread_from(context.message)
            reply.assign_trace_from(context.message)
            await responder.send_reply(reply)

        if context.settings.get("debug.monitor_ping"):
            await context.profile.notify(
                "acapy::ping::received",
                {
                    "comment": context.message.comment,
                    "connection_id": context.message_receipt.connection_id,
                    "responded": context.message.response_requested,
                    "state": "received",
                    "thread_id": context.message._thread_id,
                },
            )



from ..ld_proofs.constants import (
    CREDENTIALS_CONTEXT_V1_URL,
    CREDENTIALS_CONTEXT_V2_URL,
    DID_V1_CONTEXT_URL,
    SECURITY_CONTEXT_BBS_URL,
    SECURITY_CONTEXT_ED25519_2020_URL,
    SECURITY_CONTEXT_MULTIKEY_URL,
    SECURITY_CONTEXT_V1_URL,
    SECURITY_CONTEXT_V2_URL,
    SECURITY_CONTEXT_V3_URL,
)
from .contexts import (
    BBS_V1,
    CITIZENSHIP_V1,
    CREDENTIALS_V1,
    CREDENTIALS_V2,
    DID_V1,
    DIF_PRESENTATION_SUBMISSION_V1,
    ED25519_2020_V1,
    EXAMPLES_V1,
    MULTIKEY_V1,
    ODRL,
    SCHEMA_ORG,
    SECURITY_V1,
    SECURITY_V2,
    SECURITY_V3_UNSTABLE,
    VACCINATION_V1,
)
from .dids import (
    DID_EXAMPLE_48939859,
    DID_SOV_QqEfJxe752NCmWqR5TssZ5,
    DID_z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL,
    DID_zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq,
    DID_zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa,
)

DOCUMENTS = {
    DID_z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL.get(
        "id"
    ): DID_z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL,
    DID_zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq.get(
        "id"
    ): DID_zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq,
    DID_zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa.get(
        "id"
    ): DID_zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa,
    DID_EXAMPLE_48939859.get("id"): DID_EXAMPLE_48939859,
    DID_SOV_QqEfJxe752NCmWqR5TssZ5.get("id"): DID_SOV_QqEfJxe752NCmWqR5TssZ5,
    SECURITY_CONTEXT_V1_URL: SECURITY_V1,
    SECURITY_CONTEXT_V2_URL: SECURITY_V2,
    SECURITY_CONTEXT_V3_URL: SECURITY_V3_UNSTABLE,
    DID_V1_CONTEXT_URL: DID_V1,
    CREDENTIALS_CONTEXT_V1_URL: CREDENTIALS_V1,
    CREDENTIALS_CONTEXT_V2_URL: CREDENTIALS_V2,
    SECURITY_CONTEXT_BBS_URL: BBS_V1,
    SECURITY_CONTEXT_ED25519_2020_URL: ED25519_2020_V1,
    SECURITY_CONTEXT_MULTIKEY_URL: MULTIKEY_V1,
    "https://www.w3.org/2018/credentials/examples/v1": EXAMPLES_V1,
    "https://w3id.org/citizenship/v1": CITIZENSHIP_V1,
    "https://www.w3.org/ns/odrl.jsonld": ODRL,
    "http://schema.org/": SCHEMA_ORG,
    "https://w3id.org/vaccination/v1": VACCINATION_V1,
    "https://identity.foundation/presentation-exchange/submission/v1": DIF_PRESENTATION_SUBMISSION_V1,
}


def custom_document_loader(url: str, options: dict):
    # Check if full url (with fragments is in document map)
    if url in DOCUMENTS:
        return {
            "contentType": "application/ld+json",
            "contextUrl": None,
            "document": DOCUMENTS[url],
            "documentUrl": url,
        }

    # Otherwise look if it is present without fragment
    without_fragment = url.split("#")[0]
    if without_fragment in DOCUMENTS:
        return {
            "contentType": "application/ld+json",
            "contextUrl": None,
            "document": DOCUMENTS[without_fragment],
            "documentUrl": url,
        }

    raise Exception(f"No custom context support for {url}")



"""Represents an request for an invitation from the introduction service."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import INVITATION_REQUEST, PROTOCOL_PACKAGE

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.invitation_request_handler.InvitationRequestHandler"
)


class InvitationRequest(AgentMessage):
    """Class representing an invitation request."""

    class Meta:
        """Metadata for an invitation request."""

        handler_class = HANDLER_CLASS
        message_type = INVITATION_REQUEST
        schema_class = "InvitationRequestSchema"

    def __init__(
        self, *, responder: Optional[str] = None, message: Optional[str] = None, **kwargs
    ):
        """Initialize invitation request object.

        Args:
            responder: The name of the agent initiating the introduction
            message: Comments on the introduction
            kwargs: Additional key word arguments for the message
        """
        super().__init__(**kwargs)
        self.responder = responder
        self.message = message


class InvitationRequestSchema(AgentMessageSchema):
    """Invitation request schema class."""

    class Meta:
        """Invitation request schema metadata."""

        model_class = InvitationRequest
        unknown = EXCLUDE

    responder = fields.Str(
        required=True,
        metadata={
            "description": "Agent name initiating the introduction",
            "example": "Alice's agent",
        },
    )
    message = fields.Str(
        required=False,
        allow_none=True,
        metadata={
            "description": "Comments on the introduction",
            "example": "Hello Charlie, it's Alice; allow me to present Bob",
        },
    )



"""Message and inner object type identifiers for present-proof protocol v2.0."""

from ...didcomm_prefix import DIDCommPrefix

SPEC_URI = (
    "https://github.com/decentralized-identity/aries-rfcs/tree/"
    "560ffd23361f16a01e34ccb7dcc908ec28c5ddb1/features/0015-acks"
)

# Message types
NOTIF_10_ACK = "notification/1.0/ack"

PROTOCOL_PACKAGE = "acapy_agent.protocols.notification.v1_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        NOTIF_10_ACK: f"{PROTOCOL_PACKAGE}.messages.ack.V10Ack",
    }
)



"""Ledger utilities."""

import re

from ..core.profile import Profile

TAA_ACCEPTED_RECORD_TYPE = "taa_accepted"

DID_EVENT_PREFIX = "acapy::REGISTER_DID::"
EVENT_LISTENER_PATTERN = re.compile(f"^{DID_EVENT_PREFIX}(.*)?$")


async def notify_register_did_event(profile: Profile, did: str, meta_data: dict):
    """Send notification for a DID post-process event."""
    await profile.notify(
        DID_EVENT_PREFIX + did,
        meta_data,
    )



"""Merkel Validation Utils."""

import hashlib
from binascii import hexlify

hash_function = hashlib.sha256()

NIBBLE_TERMINATOR = 16
hti = {c: i for i, c in enumerate("0123456789abcdef")}


def sha3_256(x):
    """Return 256 bit digest."""
    return hashlib.sha3_256(x).digest()


def encode_hex(b):
    """Return bytes object for string or hexadecimal rep for bytes input.

    Args:
        b: string or bytes

    """
    if isinstance(b, str):
        b = bytes(b, "utf-8")
    if isinstance(b, bytes):
        return str(hexlify(b), "utf-8")
    raise TypeError("Value must be an instance of str or bytes")


def bin_to_nibbles(s):
    """Convert string s to nibbles (half-bytes)."""
    return [hti[c] for c in encode_hex(s)]


def unpack_to_nibbles(bindata):
    """Unpack packed binary data to nibbles.

    Args:
        bindata: binary packed from nibbles

    """

    o = bin_to_nibbles(bindata)
    flags = o[0]
    if flags & 2:
        o.append(NIBBLE_TERMINATOR)
    if flags & 1 == 1:
        o = o[1:]
    else:
        o = o[2:]
    return o


def audit_path_length(index: int, tree_size: int):
    """Return AuditPath length.

    Args:
        index: Leaf index
        tree_size: Tree size

    """
    length = 0
    last_node = tree_size - 1
    while last_node > 0:
        if index % 2 or index < last_node:
            length += 1
        index //= 2
        last_node //= 2

    return length


def ascii_chr(value):
    """Return bytes object."""
    return bytes([value])



from unittest import TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import ATTACHED_MESSAGE
from ..messages_attach import MessagesAttach


class TestConfig:
    test_author_did = "LjgpST2rjsoxYegQDRm7EL"
    test_author_verkey = "4uq1mUATWKZArwyuPgjUEw5UBysWNbkf2SN6SqVwbfSqCfnbm1Vnfw"
    test_endorser_did = "V4SGRU86Z58d6TV7PBUe6f"
    test_transaction_message = {
        "attr_names": ["first_name", "last_name"],
        "name": "test_schema",
        "version": "2.1",
    }
    test_mechanism = "manual"
    test_taaDigest = "f50feca7bd4202c2ab977006761d36bd6f23e4c6a7e0fc2feb9f62"
    test_time = 1597708800


class TestMessagesAttach(TestCase, TestConfig):
    def setUp(self):
        self.messages_attach = MessagesAttach(
            author_did=self.test_author_did,
            author_verkey=self.test_author_verkey,
            endorser_did=self.test_endorser_did,
            transaction_message=self.test_transaction_message,
            mechanism=self.test_mechanism,
            taaDigest=self.test_taaDigest,
            time=self.test_time,
        )

    def test_init(self):
        """Test initialization."""
        assert self.messages_attach.data["json"]["identifier"] == self.test_author_did
        assert self.messages_attach.data["json"]["signatures"] == {
            self.test_author_did: self.test_author_verkey
        }
        assert self.messages_attach.data["json"]["endorser"] == self.test_endorser_did
        assert (
            self.messages_attach.data["json"]["operation"]["data"]
            == self.test_transaction_message
        )
        assert (
            self.messages_attach.data["json"]["taaAcceptance"]["mechanism"]
            == self.test_mechanism
        )
        assert (
            self.messages_attach.data["json"]["taaAcceptance"]["taaDigest"]
            == self.test_taaDigest
        )
        assert (
            self.messages_attach.data["json"]["taaAcceptance"]["time"] == self.test_time
        )

    def test_type(self):
        """Test type."""
        assert self.messages_attach._type == DIDCommPrefix.qualify_current(
            ATTACHED_MESSAGE
        )

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "messages_attach.MessagesAttachSchema.load"
    )
    def test_deserialize(self, mock_messages_attach_schema_load):
        """
        Test deserialization.
        """
        obj = self.messages_attach

        messages_attach = MessagesAttach.deserialize(obj)
        mock_messages_attach_schema_load.assert_called_once_with(obj)

        assert messages_attach is mock_messages_attach_schema_load.return_value

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "messages_attach.MessagesAttachSchema.dump"
    )
    def test_serialize(self, mock_messages_attach_schema_dump):
        """
        Test serialization.
        """
        messages_attach_dict = self.messages_attach.serialize()
        mock_messages_attach_schema_dump.assert_called_once_with(self.messages_attach)

        assert messages_attach_dict is mock_messages_attach_schema_dump.return_value



"""Handler for keylist-update-response message."""

from .....core.profile import Profile
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageNotFoundError
from .....wallet.error import WalletNotFoundError
from ..manager import MediationManager
from ..messages.keylist_update_response import KeylistUpdateResponse
from ..route_manager import RouteManager


class KeylistUpdateResponseHandler(BaseHandler):
    """Handler for keylist-update-response message."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle keylist-update-response message."""
        self._logger.debug("%s called with context %s", self.__class__.__name__, context)
        assert isinstance(context.message, KeylistUpdateResponse)

        if not context.connection_ready:
            raise HandlerException("Invalid mediation request: no active connection")

        mgr = MediationManager(context.profile)
        await mgr.store_update_results(
            context.connection_record.connection_id, context.message.updated
        )
        await self.notify_keylist_updated(
            context.profile, context.connection_record.connection_id, context.message
        )

    async def notify_keylist_updated(
        self, profile: Profile, connection_id: str, response: KeylistUpdateResponse
    ):
        """Notify of keylist update response received."""
        route_manager = profile.inject(RouteManager)
        self._logger.debug(
            "Retrieving connection ID from route manager of type %s",
            type(route_manager).__name__,
        )
        try:
            key_to_connection = {
                updated.recipient_key: await route_manager.connection_from_recipient_key(
                    profile, updated.recipient_key
                )
                for updated in response.updated
            }
        except (StorageNotFoundError, WalletNotFoundError) as err:
            raise HandlerException(
                "Unknown recipient key received in keylist update response"
            ) from err

        await profile.notify(
            MediationManager.KEYLIST_UPDATED_EVENT,
            {
                "connection_id": connection_id,
                "thread_id": response._thread_id,
                "updated": [update.serialize() for update in response.updated],
                "mediated_connections": {
                    key: conn.connection_id for key, conn in key_to_connection.items()
                },
            },
        )
